"""
PyTorch backend for GPU-accelerated computations.
"""

import numpy as np
import torch
from .base import BaseBackend


class PyTorchBackend(BaseBackend):
    """PyTorch-based backend for GPU computations."""

    def __init__(self, device=None):
        if device is None:
            if torch.cuda.is_available():
                self._device = torch.device('cuda')
                self.dtype = torch.float64
                self.complex_dtype = torch.complex128
            elif hasattr(torch.backends, 'mps') and torch.backends.mps.is_available():
                self._device = torch.device('mps')
                # MPS doesn't support float64, use float32
                self.dtype = torch.float32
                self.complex_dtype = torch.complex64
            else:
                self._device = torch.device('cpu')
                self.dtype = torch.float64
                self.complex_dtype = torch.complex128
        else:
            self._device = torch.device(device)
            if 'mps' in str(device):
                self.dtype = torch.float32
                self.complex_dtype = torch.complex64
            else:
                self.dtype = torch.float64
                self.complex_dtype = torch.complex128

        print(f"PyTorch backend initialized on device: {self._device} (dtype: {self.dtype})")

    @property
    def name(self):
        return 'pytorch'

    @property
    def device(self):
        return str(self._device)

    def to_array(self, data, dtype=None):
        if dtype is None:
            dtype = self.dtype
        if isinstance(data, torch.Tensor):
            return data.to(device=self._device, dtype=dtype)
        return torch.tensor(data, dtype=dtype, device=self._device)

    def to_numpy(self, data):
        if isinstance(data, torch.Tensor):
            return data.detach().cpu().numpy()
        return np.asarray(data)

    def zeros(self, shape, dtype=None):
        if dtype is None:
            dtype = self.dtype
        return torch.zeros(shape, dtype=dtype, device=self._device)

    def ones(self, shape, dtype=None):
        if dtype is None:
            dtype = self.dtype
        return torch.ones(shape, dtype=dtype, device=self._device)

    def eye(self, n, dtype=None):
        if dtype is None:
            dtype = self.dtype
        return torch.eye(n, dtype=dtype, device=self._device)

    def diag(self, v):
        if not isinstance(v, torch.Tensor):
            v = self.to_array(v)
        return torch.diag(v)

    def dot(self, a, b):
        if not isinstance(a, torch.Tensor):
            a = self.to_array(a)
        if not isinstance(b, torch.Tensor):
            b = self.to_array(b)
        if a.ndim == 1 and b.ndim == 1:
            return torch.dot(a, b)
        return torch.matmul(a, b)

    def matmul(self, a, b):
        if not isinstance(a, torch.Tensor):
            a = self.to_array(a)
        if not isinstance(b, torch.Tensor):
            b = self.to_array(b)
        return torch.matmul(a, b)

    def inv(self, a):
        if not isinstance(a, torch.Tensor):
            a = self.to_array(a)
        return torch.linalg.inv(a)

    def det(self, a):
        if not isinstance(a, torch.Tensor):
            a = self.to_array(a)
        return torch.linalg.det(a)

    def eig(self, a):
        if not isinstance(a, torch.Tensor):
            a = self.to_array(a)
        return torch.linalg.eig(a)

    def sqrt(self, x):
        if not isinstance(x, torch.Tensor):
            x = self.to_array(x)
        return torch.sqrt(x)

    def exp(self, x):
        if not isinstance(x, torch.Tensor):
            x = self.to_array(x)
        return torch.exp(x)

    def sin(self, x):
        if not isinstance(x, torch.Tensor):
            x = self.to_array(x)
        return torch.sin(x)

    def cos(self, x):
        if not isinstance(x, torch.Tensor):
            x = self.to_array(x)
        return torch.cos(x)

    def tan(self, x):
        if not isinstance(x, torch.Tensor):
            x = self.to_array(x)
        return torch.tan(x)

    def arcsin(self, x):
        if not isinstance(x, torch.Tensor):
            x = self.to_array(x)
        return torch.asin(x)

    def arccos(self, x):
        if not isinstance(x, torch.Tensor):
            x = self.to_array(x)
        return torch.acos(x)

    def arctan2(self, y, x):
        if not isinstance(y, torch.Tensor):
            y = self.to_array(y)
        if not isinstance(x, torch.Tensor):
            x = self.to_array(x)
        return torch.atan2(y, x)

    def abs(self, x):
        if not isinstance(x, torch.Tensor):
            x = self.to_array(x)
        return torch.abs(x)

    def log(self, x):
        if not isinstance(x, torch.Tensor):
            x = self.to_array(x)
        return torch.log(x)

    def sum(self, x, axis=None):
        if not isinstance(x, torch.Tensor):
            x = self.to_array(x)
        if axis is None:
            return torch.sum(x)
        return torch.sum(x, dim=axis)

    def stack(self, arrays, axis=0):
        tensors = [self.to_array(a) if not isinstance(a, torch.Tensor) else a for a in arrays]
        return torch.stack(tensors, dim=axis)

    def concatenate(self, arrays, axis=0):
        tensors = [self.to_array(a) if not isinstance(a, torch.Tensor) else a for a in arrays]
        return torch.cat(tensors, dim=axis)

    def transpose(self, a, axes=None):
        if not isinstance(a, torch.Tensor):
            a = self.to_array(a)
        if axes is None:
            return a.T
        return a.permute(axes)

    def reshape(self, a, shape):
        if not isinstance(a, torch.Tensor):
            a = self.to_array(a)
        return a.reshape(shape)

    def linspace(self, start, stop, num):
        return torch.linspace(start, stop, num, dtype=self.dtype, device=self._device)

    def arange(self, start, stop=None, step=1):
        if stop is None:
            return torch.arange(start, dtype=self.dtype, device=self._device)
        return torch.arange(start, stop, step, dtype=self.dtype, device=self._device)

    def meshgrid(self, *xi, indexing='xy'):
        tensors = [self.to_array(x) if not isinstance(x, torch.Tensor) else x for x in xi]
        return torch.meshgrid(*tensors, indexing=indexing)

    def einsum(self, subscripts, *operands):
        tensors = [self.to_array(op) if not isinstance(op, torch.Tensor) else op for op in operands]
        return torch.einsum(subscripts, *tensors)

    def block_diag(self, *matrices):
        tensors = [self.to_array(m) if not isinstance(m, torch.Tensor) else m for m in matrices]
        return torch.block_diag(*tensors)

    # Additional PyTorch-specific methods
    def sign(self, x):
        if not isinstance(x, torch.Tensor):
            x = self.to_array(x)
        return torch.sign(x)

    def size(self, x):
        if isinstance(x, torch.Tensor):
            return x.numel()
        return np.size(x)

    def copy(self, x):
        if isinstance(x, torch.Tensor):
            return x.clone()
        return self.to_array(x)

    def clip(self, x, a_min, a_max):
        if not isinstance(x, torch.Tensor):
            x = self.to_array(x)
        return torch.clamp(x, a_min, a_max)

    def where(self, condition, x, y):
        if not isinstance(condition, torch.Tensor):
            condition = self.to_array(condition, dtype=torch.bool)
        if not isinstance(x, torch.Tensor):
            x = self.to_array(x)
        if not isinstance(y, torch.Tensor):
            y = self.to_array(y)
        return torch.where(condition, x, y)

    def maximum(self, x, y):
        if not isinstance(x, torch.Tensor):
            x = self.to_array(x)
        if not isinstance(y, torch.Tensor):
            y = self.to_array(y)
        return torch.maximum(x, y)

    def minimum(self, x, y):
        if not isinstance(x, torch.Tensor):
            x = self.to_array(x)
        if not isinstance(y, torch.Tensor):
            y = self.to_array(y)
        return torch.minimum(x, y)

    # Batch operations for resolution calculations
    def batch_inv(self, a):
        """Batch matrix inversion for shape (n, m, m)."""
        if not isinstance(a, torch.Tensor):
            a = self.to_array(a)
        return torch.linalg.inv(a)

    def batch_det(self, a):
        """Batch determinant for shape (n, m, m)."""
        if not isinstance(a, torch.Tensor):
            a = self.to_array(a)
        return torch.linalg.det(a)

    def batch_matmul(self, a, b):
        """Batch matrix multiplication for shapes (n, i, j) @ (n, j, k)."""
        if not isinstance(a, torch.Tensor):
            a = self.to_array(a)
        if not isinstance(b, torch.Tensor):
            b = self.to_array(b)
        return torch.bmm(a, b)

    def batch_similarity_transform(self, A, B):
        """Batch similarity transform: A @ B @ A.T for batched matrices."""
        if not isinstance(A, torch.Tensor):
            A = self.to_array(A)
        if not isinstance(B, torch.Tensor):
            B = self.to_array(B)
        return torch.bmm(torch.bmm(A, B), A.transpose(-2, -1))



import numpy as N
import scipy
import scipy.integrate as integrate
import lattice_calculator
import rescalc
import copy

pi=N.pi
def ConvRes(sqw,pref,H,K,L,W,myrescal,setup,p,METHOD='fixed',ACCURACY=[7,0]):
    """"Numerically calculate the convolution of a user-defined cross-section
    function with the Cooper-Nathans resolution function for a 3-axis neutron
    scattering experiment. See manual for details."""

    #Get standard orthonormnal coordinate system, cell parameters and reciprocal cell parameters...
    xvec=myrescal.lattice_calculator.x
    yvec=myrescal.lattice_calculator.y
    zvec=myrescal.lattice_calculator.z
    #print 'xvec ',xvec[0,0]
    #print 'xvec ',xvec[1,0]
    #print 'yvec ',yvec
    #print 'zvec ',zvec
    npts=H.shape[0]
    #Calculate the resolution matrix...
    R0,RMS=myrescal.ResMatS(H,K,L,W,setup);
    #print 'shape ', RMS.shape
    #print RMS[3,3,:]
    #print type(RMS)
    Mzz=N.array(RMS[3,3,:])
    Mww=N.array(RMS[2,2,:])
    Mxx=N.array(RMS[0,0,:])
    Mxy=N.array(RMS[0,1,:])
    Mxw=N.array(RMS[0,2,:])
    Myy=N.array(RMS[1,1,:])
    Myw=N.array(RMS[1,2,:])
    Mxx=Mxx-Mxw**2/Mww
    Mxy=Mxy-Mxw*Myw/Mww
    Myy=Myy-Myw**2/Mww
    MMxx=Mxx-Mxy**2/Myy
    detM=MMxx*Myy*Mzz*Mww


    tqz=1./N.sqrt(Mzz); tqx=1/N.sqrt(MMxx)
    tqyy=1./N.sqrt(Myy); tqyx=-Mxy/Myy/N.sqrt(MMxx)
    tqww=1./N.sqrt(Mww); tqwy=-Myw/Mww/N.sqrt(Myy)
    tqwx=-(Mxw/Mww-Myw/Mww*Mxy/Myy)/N.sqrt(MMxx)

##    print 'Mzz ',Mzz
##    print 'Mww ',Mww
##    print 'Mxx ',Mxx
##    print 'Mxy ',Mxy
##    print 'Mxw ',Mxw
##    print 'Myy ',Myy
##    print 'Myw ',Myw
##    print 'detM ', detM
##    print 'tqz ',tqz
##    print 'tqx ', tqx
##    print 'tqyy ',tqyy
##    print 'tqyx ',tqyx
##    print 'tqww ',tqww
##    print 'tqwy ',tqwy
##    print 'tqwx ',tqwx


    #================================================================================================
    myargs=(H,K,L,W,p)
    testsqw=sqw(H,K,L,W,p)
    modes=testsqw.shape[0]
    try:
        (modes,points)=testsqw.shape
    except:
        (modes,)=testsqw.shape
        points=1
    if pref==None:
       prefactor=N.ones((modes,points),'float64')
       bgr=0
    else:
        prefactor,bgr=pref(H,K,L,W,myrescal,p)

#    %========================= fix  method ===========================================================
    if METHOD=='fixed':
       M=ACCURACY
       step1=pi/(2*M[0]+1); step2=pi/(2*M[1]+1)
       dd1=N.linspace(-pi/2+step1/2,pi/2-step1/2,(2*M[0]+1))
       dd2=N.linspace(-pi/2+step2/2,pi/2-step2/2,(2*M[1]+1))
       convs=N.zeros((modes,npts),'float64')
       conv=N.zeros((npts,),'float64')
       #cx,cy,cw=N.mgrid(dd1,dd1,dd1)
       a=-pi/2+step1/2
       b=pi/2-step1/2
       dpts=N.complex(0,2*M[0]+1)
       cx,cy,cw=N.mgrid[a:b:dpts,a:b:dpts,a:b:dpts]
       #tx=N.tan(N.reshape(cx,(N.size(cx),)))
       tx=N.tan(cx.reshape(N.size(cx),1,order='F').copy().T).flatten()
       ty=N.tan(cy.reshape(N.size(cx),1,order='F').copy().T).flatten()
       tw=N.tan(cw.reshape(N.size(cx),1,order='F').copy().T).flatten()
       tz=N.tan(dd2)
       norm=(1+tx**2)*(1+ty**2)*(1+tw**2)*N.exp(-0.5*(tx**2+ty**2))*N.exp(-0.5*(tw**2))
       normz=N.exp(-0.5*(tz**2))*(1+tz**2)
       for iz in range(N.size(tz)):
           for i in range(npts):
              dQ1=tqx[i]*tx
              dQ2=tqyy[i]*ty+tqyx[i]*tx
              dW=tqwx[i]*tx+tqwy[i]*ty+tqww[i]*tw
              dQ4=tqz[i]*tz[iz]
              H1=H[i]+dQ1*xvec[0,i]+dQ2*yvec[0,i]+dQ4*zvec[0,i]
              K1=K[i]+dQ1*xvec[1,i]+dQ2*yvec[1,i]+dQ4*zvec[1,i]
              L1=L[i]+dQ1*xvec[2,i]+dQ2*yvec[2,i]+dQ4*zvec[2,i]
              W1=W[i]+dW;
              myint=sqw(H1,K1,L1,W1,p)
              for j in range(modes):
                  add=myint[j,:]*norm*normz[iz]
                  convs[j,i]=convs[j,i]+add.sum()
              #print 'conv ',conv.shape
              conv[i]=(convs[:,i]*prefactor[:,i]).sum()
       factor=step1**3*step2/N.sqrt(detM)
       #print 'factor ',factor
       conv=conv*factor
       if M[1]==0:
            conv=conv*0.79788
       if M[0]==0:
            conv=conv*0.79788**3

##    %========================= mc  method =============================================================
##
##    if strcmp(METHOD,'mc')
##       found=1;
##       if isempty(ACCURACY)
##          ACCURACY=10;
##       end;
##       M=ACCURACY;
##       convs=zeros(modes,len);
##       conv=zeros(1,len);
##       for i=1:len
##       for MonteCarlo=1:M
##          r=randn(4,1000)*pi-pi/2;
##          cx=r(1,:); cy=r(2,:); cz=r(3,:);cw=r(4,:);
##          tx=tan(cx(1:end)); ty=tan(cy(1:end)); tz=tan(cz(1:end)); tw=tan(cw(1:end));
##          norm=exp(-0.5*(tx.^2+ty.^2+tz.^2+tw.^2)).*(1+tx.^2).*(1+ty.^2).*(1+tz.^2).*(1+tw.^2);
##          dQ1=tqx(i)*tx;
##          dQ2=tqyy(i)*ty+tqyx(i)*tx;
##          dW=tqwx(i)*tx+tqwy(i)*ty+tqww(i)*tw;
##          dQ4=tqz(i)*tz;
##          H1=H(i)+dQ1*xvec(1,i)+dQ2*yvec(1,i)+dQ4*zvec(1,i);
##          K1=K(i)+dQ1*xvec(2,i)+dQ2*yvec(2,i)+dQ4*zvec(2,i);
##          L1=L(i)+dQ1*xvec(3,i)+dQ2*yvec(3,i)+dQ4*zvec(3,i);
##          W1=W(i)+dW;
##          int=feval(sqw,H1,K1,L1,W1,p);
##          for j=1:modes
##              add=int(j,:).*norm;
##              convs(j,i)=convs(j,i)+sum(add);
##          end
##          conv(i)=sum( convs(:,i).*prefactor(:,i) );
##      end;end;
##       conv=conv/M/1000*pi^4./sqrt(detM);
##    end;
##    %==================================================================================================
##
##    if found==0:
##       error('??? Error in ConvRes: Unknown convolution method! Valid options are: "fix",  "mc".');

    conv=conv*R0
    conv=conv+bgr
    return conv





import numpy as N
import pylab
import math
import unittest
from matplotlib.patches import Ellipse
import lattice_calculator
import readncnr2 as readncnr
import scriptutil as SU
import re
import simple_combine
import sys
#sys.path.append(r'c:\tripleaxisproject2\polarization')
sys.path.append(r'c:\mytripleaxisproject\polarization')
import polcorrect_lib
from matplotlib.ticker import NullFormatter, MultipleLocator
from matplotlib.ticker import FormatStrFormatter
from matplotlib.ticker import MaxNLocator
import scipy.sandbox.delaunay as D
#import numpy.core.ma as ma
import matplotlib.numerix.ma as ma
from scipy.signal.signaltools import convolve2d
import readicp



eps=1e-3
pi=N.pi


def autovectorized(f):
     """Function decorator to do vectorization only as necessary.
     vectorized functions fail for scalar inputs."""
     def wrapper(input):
         if N.isscalar(input)==False:
             return N.vectorize(f)(input)
         return f(input)
     return wrapper



@autovectorized
def myradians(x):
    return math.radians(x)

def readfiles(mydirectory,myfilebase,myend):
    myfilebaseglob=myfilebase+'*.'+myend
    print myfilebaseglob
    flist = SU.ffind(mydirectory, shellglobs=(myfilebaseglob,))
    Qx=N.array([])
    Qy=N.array([])
    Qz=N.array([])
    Counts=N.array([])
    for currfile in flist:
        print currfile
        mydata=mydatareader.readbuffer(currfile)
        Qx=N.concatenate((Qx,N.array(mydata.data['qx'])))
        Qy=N.concatenate((Qy,N.array(mydata.data['qy'])))
        Qz=N.concatenate((Qz,N.array(mydata.data['qz'])))
        Counts=N.concatenate((Counts,N.array(mydata.data['counts'])))


def plot_nodes(tri):
    for nodes in tri.triangle_nodes:
        D.fill(x[nodes],y[nodes],'b')
    pylab.show()

def plot_data(xa,ya,za,fig,nfig,colorflag=False):

    cmap = pylab.cm.jet
    cmap.set_bad('w', 1.0)
    myfilter=N.array([[0.1,0.2,0.1],[0.2,0.8,0.2],[0.1,0.2,0.1]],'d') /2.0
    zout=convolve2d(za,myfilter,mode='same')
    zima = ma.masked_where(N.isnan(zout),zout)


    ax=fig.add_subplot(3,2,nfig)
    pc=ax.pcolormesh(xa,ya,zima,shading='interp',cmap=cmap)  # working good!
#    pc=ax.imshow(zima,interpolation='bilinear',cmap=cmap)
    pc.set_clim(0.0,660.0)



    if colorflag:
        g=pylab.colorbar(pc,ticks=N.arange(0,675,100))
        print g
        #g.ticks=None
        #gax.yaxis.set_major_locator(MultipleLocator(40))
        #g.ticks(N.array([0,20,40,60,80]))

    return ax,g

def prep_data(filename):
#    Data=pylab.load(r'c:\resolution_stuff\1p4K.iexy')
    Data=pylab.load(filename)
    xt=Data[:,2]
    yt=Data[:,3]
    zorigt=Data[:,0]
    x=xt[:,zorigt>0.0]
    y=yt[:,zorigt>0.0]
    z=zorigt[:,zorigt>0.0]
#    zorig=ma.array(zorigt)
    print 'reached'
    threshold=0.0;
#    print zorigt < threshold
#    print N.isnan(zorigt)
#    z = ma.masked_where(zorigt < threshold , zorigt)
    print 'where masked ', z.shape
#should be commented out--just for testing
##    x = pylab.randn(Nu)/aspect
##    y = pylab.randn(Nu)
##    z = pylab.rand(Nu)
##    print x.shape
##    print y.shape
    # Grid
    xi, yi = N.mgrid[-5:5:100j,-5:5:100j]
    xi,yi=N.mgrid[x.min():x.max():.05,y.min():y.max():.05]
    # triangulate data
    tri = D.Triangulation(x,y)
    print 'before interpolator'
    # interpolate data
    interp = tri.nn_interpolator(z)
    print 'interpolator reached'
    zi = interp(xi,yi)
    # or, all in one line
    #    zi = Triangulation(x,y).nn_interpolator(z)(xi,yi)
#    return x,y,z
    return xi,yi,zi





def prep_data2(xt,yt,zorigt):
#    Data=pylab.load(r'c:\resolution_stuff\1p4K.iexy')
    #Data=pylab.load(filename)
    #xt=Data[:,2]
    #yt=Data[:,3]
    #zorigt=Data[:,0]
    x=xt[:,zorigt>0.0]
    y=yt[:,zorigt>0.0]
    z=zorigt[:,zorigt>0.0]
#    zorig=ma.array(zorigt)
    print 'reached'
    threshold=0.0;
#    print zorigt < threshold
#    print N.isnan(zorigt)
#    z = ma.masked_where(zorigt < threshold , zorigt)
    print 'where masked ', z.shape
#should be commented out--just for testing
##    x = pylab.randn(Nu)/aspect
##    y = pylab.randn(Nu)
##    z = pylab.rand(Nu)
##    print x.shape
##    print y.shape
    # Grid
    xi, yi = N.mgrid[-5:5:100j,-5:5:100j]
    xi,yi=N.mgrid[x.min():x.max():.001,y.min():y.max():.001]
    # triangulate data
    tri = D.Triangulation(x,y)
    print 'before interpolator'
    # interpolate data
    interp = tri.nn_interpolator(z)
    print 'interpolator reached'
    zi = interp(xi,yi)
    # or, all in one line
    #    zi = Triangulation(x,y).nn_interpolator(z)(xi,yi)
#    return x,y,z
    return xi,yi,zi



def readmeshfiles(mydirectory,myfilebase,myend):
    myfilebaseglob=myfilebase+'*.'+myend
    print myfilebaseglob
    flist = SU.ffind(mydirectory, shellglobs=(myfilebaseglob,))
    #SU.printr(flist)
    mydatareader=readicp.datareader()
    Qx=N.array([])
    Qy=N.array([])
    Qz=N.array([])
    Counts=N.array([])
    for currfile in flist:
        print currfile
        mydata=mydatareader.readbuffer(currfile)
        Qx=N.concatenate((Qx,N.array(mydata.data['Qx'])))
        Qy=N.concatenate((Qy,N.array(mydata.data['Qy'])))
        Qz=N.concatenate((Qz,N.array(mydata.data['Qz'])))
        Counts=N.concatenate((Counts,N.array(mydata.data['Counts'])))
    xa,ya,za=prep_data2(Qx,Qy,Counts);
    return xa,ya,za

def strangezone(fig):
    Nu = 10000
    aspect = 1.0
    mydirectory=r'c:\bifeo3xtal\dec7_2007'
    myfilebase='cmesh'
    myend='bt9'
    xc,yc,zc=readmeshfiles(mydirectory,'cmesh',myend) #Rm temp
    xd,yd,zd=readmeshfiles(mydirectory,'dmesh',myend) #0
    xe,ye,ze=readmeshfiles(mydirectory,'emesh',myend) #-1.3
    xf,yf,zf=readmeshfiles(mydirectory,'fmesh',myend) #0
    xg,yg,zg=readmeshfiles(mydirectory,'gmesh',myend) #1.3
    xh,yh,zh=readmeshfiles(mydirectory,'hmesh',myend) #0
    xi,yi,zi=readmeshfiles(mydirectory,'imesh',myend) #-1.3
    xj,yj,zj=readmeshfiles(mydirectory,'jmesh',myend) #0
    print 'matplotlib'

    if 1:

        #fig=pylab.figure(figsize=(8,8))
        ylim=(.485,.515)
        xlim=(.485,.515)
        #ylabel='E (meV)'
        #xlabel=r'Q$ \ \ (\AA^{-1}$)'
        ylabel='(1 1 0)'
        xlabel='(1 -1 -2)'


    if 1:
        ax,g=plot_data(xd,yd,zd,fig,4,colorflag=True)
        #ax.text(.98,.20,'E=0 KV',fontsize=14,horizontalalignment='right',verticalalignment='top',transform=ax.transAxes,color='white')
        ax.set_ylabel(ylabel)
        ax.set_xlabel(xlabel)
        ax.xaxis.set_major_formatter(NullFormatter())
        ax.set_ylim(ylim); ax.set_xlim(xlim)
        #g.ax.ticks=N.arange(0,100,20)

    if 0:
        ax,g=plot_data(xe,ye,ze,fig,2,colorflag=True)
        ax.text(.98,.20,'E=-1.3 KV',fontsize=14,horizontalalignment='right',verticalalignment='top',transform=ax.transAxes,color='white')
        ax.yaxis.set_major_formatter(NullFormatter())
        ax.xaxis.set_major_formatter(NullFormatter())
        ax.set_ylim(ylim); ax.set_xlim(xlim)

    if 0:
        ax,g=plot_data(xf,yf,zf,fig,3,colorflag=True)
        ax.text(.98,.20,'0 KV',fontsize=14,horizontalalignment='right',verticalalignment='top',transform=ax.transAxes,color='white')
        ax.set_ylabel(ylabel)
        ax.xaxis.set_major_formatter(NullFormatter())
        ax.set_ylim(ylim); ax.set_xlim(xlim)
    if 0:
        ax,g=plot_data(xg,yg,zg,fig,4,colorflag=True)
        ax.text(.98,.20,'1.3 KV',fontsize=14,horizontalalignment='right',verticalalignment='top',transform=ax.transAxes,color='white')
        ax.yaxis.set_major_locator(MultipleLocator(1))
        ax.yaxis.set_major_formatter(NullFormatter())
        ax.xaxis.set_major_formatter(NullFormatter())
        ax.set_ylim(ylim); ax.set_xlim(xlim)
    if 0:
        ax,g=plot_data(xh,yh,zh,fig,5,colorflag=True)
        ax.text(.98,.20,'0 KV',fontsize=14,horizontalalignment='right',verticalalignment='top',transform=ax.transAxes,color='white')
        ax.set_ylabel(ylabel)
        ax.set_xlabel(xlabel)
        ax.xaxis.set_minor_formatter(NullFormatter())
        ax.set_ylim(ylim); ax.set_xlim(xlim)
        fmt = FormatStrFormatter('%0.3g')  # or whatever
        ax.xaxis.set_major_formatter(fmt)
        ax.xaxis.set_major_locator(MaxNLocator(5))
    if 0:
        ax,g=plot_data(xi,yi,zi,fig,6,colorflag=True)
        ax.text(.98,.20,'-1.3 KV',fontsize=14,horizontalalignment='right',verticalalignment='top',transform=ax.transAxes,color='white')
        ax.set_xlabel(xlabel)
        ax.yaxis.set_major_formatter(NullFormatter())
        ax.xaxis.set_minor_formatter(NullFormatter())
        ax.set_ylim(ylim); ax.set_xlim(xlim)
        fmt = FormatStrFormatter('%0.3g')  # or whatever
        ax.xaxis.set_major_formatter(fmt)
        ax.xaxis.set_major_locator(MaxNLocator(5))

    if 0:
        ax,g=plot_data(xf,yf,zf,fig,6,colorflag=True)
        ax.text(.98,.20,'0 KV',fontsize=14,horizontalalignment='right',verticalalignment='top',transform=ax.transAxes,color='white')
        ax.set_xlabel(xlabel)
        ax.yaxis.set_major_formatter(NullFormatter())
        ax.set_ylim(ylim); ax.set_xlim(xlim)
        #fmt = FormatStrFormatter('%1.4g')  # or whatever
        #ax.yaxis.set_major_formatter(fmt)


    if 1:
        print 'gca ', fig.gca()
        for im in fig.gca().get_images():
            print im
            im.set_clim(0.0,660.0)
        #pylab.show()
    if 0:
        print 'saving'
        pylab.savefig(r'c:\sqltest\demo.pdf',dpi=150)
        print 'saved'
    if 0:
        pylab.show()



def cartoon():
    if 1:
        #h k 0
        ax=pylab.subplot(3,4,1)
        #1 1 0
        x1=N.array([1])
        y1=N.array([1])

        pylab.plot(x1,y1,'bo',markersize=10,markerfacecolor='blue',markeredgecolor='blue')
        #-1 -1 0
        x1=N.array([-1])
        y1=N.array([-1])
        pylab.plot(x1,y1,'bo',markersize=10,markerfacecolor='blue',markeredgecolor='blue')
        #-1 2 0
        x1=N.array([-1])
        y1=N.array([2])
        pylab.plot(x1,y1,'bo',markersize=10,markerfacecolor='blue',markeredgecolor='blue')
        #1 -2 0
        x1=N.array([1])
        y1=N.array([-2])
        pylab.plot(x1,y1,'bo',markersize=10,markerfacecolor='blue',markeredgecolor='blue')
        #2 -1 0
        x1=N.array([2])
        y1=N.array([-1])
        pylab.plot(x1,y1,'bo',markersize=10,markerfacecolor='blue',markeredgecolor='blue')
        #-2 1 0
        x1=N.array([-2])
        y1=N.array([1])
        pylab.plot(x1,y1,'bo',markersize=10,markerfacecolor='blue',markeredgecolor='blue')
        pylab.xlabel('[1 0 0]')
        pylab.ylabel('[0 1 0]')
        #draw vertical line
        #x1=[0,0]
        #y1=[-2.5,2.5]
        #pylab.plot(x1,y1,linewidth=3.0)
        #draw horizontal line
        #x1=[-2.5,2.5]
        #y1=[0,0]
        #pylab.plot(x1,y1,linewidth=3.0,color='blue')

        #draw line 1 1 0
        x1=[1,-1]
        y1=[1,-1]
        pylab.plot(x1,y1,linewidth=3.0,color='blue')

        #draw line 1 2 0
        x1=[1,-1]
        y1=[-2,2]
        pylab.plot(x1,y1,linewidth=3.0,color='blue')

        #draw line 1 1 0
        x1=[-2,2]
        y1=[1,-1]
        pylab.plot(x1,y1,linewidth=3.0,color='blue')



        s=r'$\delta 2\bar{\delta} 0$'
        pylab.text(1.2,-2,s,fontsize=10)

        s=r'$2\delta \bar{\delta} 0$'
        pylab.text(1.8,-0.7,s,fontsize=10)

        s=r'$\delta \delta 0$'
        pylab.text(1,1.2,s,fontsize=10)

        s=r'$\bar{\delta} \bar{\delta} 0$'
        pylab.text(-1,-1.6,s,fontsize=10)

        s=r'$2\bar{\delta} \delta 0$'
        pylab.text(-2,1.2,s,fontsize=10)

        s=r'$\bar{\delta} 2\delta 0$'
        pylab.text(-.7,2.0,s,fontsize=10)

        pylab.axis([-3.5,3.5,-3.5,3.5])
        ax.yaxis.set_major_formatter(NullFormatter())
        ax.xaxis.set_major_formatter(NullFormatter())

    if 1:
        #h h l
        ax=pylab.subplot(3,4,2)

        # 1 1 0
        x1=N.array([4])
        y1=N.array([0])
        pylab.plot(x1,y1,'bo',markersize=10,markerfacecolor='blue',markeredgecolor='blue')
        s=r'$\delta \delta 0$'
        pylab.text(4.8,0.0,s,fontsize=10)

        # -1 -1 0
        x1=N.array([-4])
        y1=N.array([0])
        pylab.plot(x1,y1,'bo',markersize=10,markerfacecolor='blue',markeredgecolor='blue')
        s=r'$\bar{\delta} \bar{\delta} 0$'
        pylab.text(-6.8,0.0,s,fontsize=10)

        # -1 2 0
        x1=N.array([2])
        y1=N.array([0])
        pylab.plot(x1,y1,'bo',markersize=5,markerfacecolor='white')
        s=r'$\bar{\delta} 2\delta 0$'
        pylab.text(0.5,0.4,s,fontsize=10)

        # 2 -1 0
        #x1=N.array([2])
        #y1=N.array([0])
        #pylab.plot(x1,y1,'bo',markersize=10)
        s=r'$2\delta \bar{\delta} 0$'
        pylab.text(0.5,-1.4,s,fontsize=10)


        # -2 1 0
        x1=N.array([-2])
        y1=N.array([0])
        pylab.plot(x1,y1,'bo',markersize=5,markerfacecolor='white')
        s=r'$2\bar{\delta} \delta 0$'
        pylab.text(-3.5,0.4,s,fontsize=10)

        # 1 -2 0
        #x1=N.array([2])
        #y1=N.array([0])
        #pylab.plot(x1,y1,'bo',markersize=10)
        s=r'$\delta 2\bar{\delta} 0$'
        pylab.text(-3.5,-1.4,s,fontsize=10)



        pylab.axis([-8.0,8.0,-3.5,3.5])
        ax.yaxis.set_major_formatter(NullFormatter())
        ax.xaxis.set_major_formatter(NullFormatter())

        pylab.xlabel('[1 1 0]')
        pylab.ylabel('[0 0 1]')

        #ax=pylab.subplot(4,2,2)
    if 1:
        #h k h-k
        ax=pylab.subplot(3,2,2)
        # 1 1 0
        x1=N.array([4])
        y1=N.array([0])
        pylab.plot(x1,y1,'bo',markersize=20,markerfacecolor='blue',markeredgecolor='blue')
        s=r'$\delta \delta 0$'
        pylab.text(4.8,0.0,s,fontsize=20)

        # -1 -1 0
        x1=N.array([-4])
        y1=N.array([0])
        pylab.plot(x1,y1,'bo',markersize=20,markerfacecolor='blue',markeredgecolor='blue')
        s=r'$\bar{\delta} \bar{\delta} 0$'
        pylab.text(-6.8,0.0,s,fontsize=20)

        # -1 2 0
        x1=N.array([2])
        y1=N.array([4])
        pylab.plot(x1,y1,'bo',markersize=5,markerfacecolor='white')
        s=r'$\bar{\delta} 2\delta 0$'
        pylab.text(0.5,4.4,s,fontsize=20)


        # -2 1 0
        x1=N.array([-2])
        y1=N.array([4])
        pylab.plot(x1,y1,'bo',markersize=10,markerfacecolor='white')
        s=r'$2\bar{\delta} \delta 0$'
        pylab.text(-3.0,4.4,s,fontsize=20)


        # 2 -1 0
        x1=N.array([2])
        y1=N.array([-4])
        pylab.plot(x1,y1,'bo',markersize=10,markerfacecolor='gray')
        s=r'$2\delta \bar{\delta} 0$'
        pylab.text(0.5,-6.8,s,fontsize=20)


        # 1 -2 0
        x1=N.array([-2])
        y1=N.array([-4])
        pylab.plot(x1,y1,'bo',markersize=5,markerfacecolor='gray')
        s=r'$\delta 2\bar{\delta} 0$'
        pylab.text(-3.0,-6.8,s,fontsize=20)


        pylab.axis([-10.5,10.5,-10.5,10.5])
        ax.yaxis.set_major_formatter(NullFormatter())
        ax.xaxis.set_major_formatter(NullFormatter())

        pylab.xlabel('[1 1 0]')
        pylab.ylabel('[-1 1 -2]')

    if 0:
        pylab.subplot(3,2,3)
        pylab.subplot(3,2,4)
        pylab.subplot(3,2,5)
        pylab.subplot(3,2,6)
        pylab.subplots_adjust(wspace=0.5,hspace=0.6)
    return


if __name__=="__main__":
    if 1:
        fig=pylab.figure(figsize=(8,8))
        a=N.array([5.5436],'d')
        b=N.array([5.5436],'d')
        c=N.array([13.8844],'d')
        alpha=N.array([90],'d')
        beta=N.array([90],'d')
        gamma=N.array([120],'d')
 #       orient1=N.array([[0,1,1]],'d')
        orient1=N.array([[0,0,1]],'d')
        orient2=N.array([[1,1,0]],'d')
        mylattice=lattice_calculator.lattice(a=a,b=b,c=c,alpha=alpha,beta=beta,gamma=gamma,\
                               orient1=orient1,orient2=orient2)

        EXP={}
        EXP['ana']={}
        EXP['ana']['tau']='pg(002)'
        EXP['mono']={}
        EXP['mono']['tau']='pg(002)';
        EXP['ana']['mosaic']=25
        EXP['mono']['mosaic']=25
        EXP['sample']={}
        EXP['sample']['mosaic']=25
        EXP['sample']['vmosaic']=25
        EXP['hcol']=N.array([40, 40, 40, 80],'d')
        EXP['vcol']=N.array([120, 120, 120, 120],'d')
        EXP['infix']=-1 #positive for fixed incident energy
        EXP['efixed']=14.7
        EXP['method']=0
        setup=[EXP]

        mydirectory=r'c:\BiFeO3xtal\Oct9_2007'
        myfilebase='bfo_spinflip_he3flip51585'
        myfilebase2='bfo_spinflip51579'
        myfilebase3='bfo_spinflip51583'
        myend='bt7'
        data={}
        #pm
        myfilebaseglob=myfilebase+'*.'+myend
        print myfilebaseglob
        flist = SU.ffind(mydirectory, shellglobs=(myfilebaseglob,))
        myfilestr=flist[0]
        Counts=N.array([],'float64')
        monlist=[]
        timestamp={}
        mydatareader=readncnr.datareader()
        mydata=mydatareader.readbuffer(myfilestr)
        S1=N.array(mydata.data['a3'][1:])
        S2=N.array(mydata.data['a4'][1:])
        A2=N.array(mydata.data['a5'][1:])*2
        M2=N.array(mydata.data['a2'][1:])
        Counts_pm=N.concatenate((Counts,N.array(mydata.data['detector'][1:])))
        H_pm,K_pm,L_pm,E_pm,Q_pm,Ei_pm,Ef_pm=mylattice.SpecWhere(myradians(M2),myradians(S1),myradians(S2),myradians(A2),setup)
        monlist.append(mydata.data['monitor'][0])
        Counts_pm=Counts_pm[:-1]
        H_pm=H_pm[:-1]
        K_pm=K_pm[:-1]
        L_pm=L_pm[:-1]
        E_pm=E_pm[:-1]
        Q_pm=Q_pm[:-1]
        Ei_pm=Ei_pm[:-1]
        Ef_pm=Ef_pm[:-1]
        timestamp['pm']=N.array(mydata.data['timestamp'][1:])[:-1]
        #print H_pm
        #begin mp
        myfilebaseglob=myfilebase2+'*.'+myend
        print myfilebaseglob
        flist = SU.ffind(mydirectory, shellglobs=(myfilebaseglob,))
        myfilebaseglob=myfilebase3+'*.'+myend
        flist2 = SU.ffind(mydirectory, shellglobs=(myfilebaseglob,))

        myfilestr=flist[0]
        Counts=N.array([],'float64')
        mydatareader=readncnr.datareader()
        mydata1=mydatareader.readbuffer(myfilestr)
        S1=N.array(mydata1.data['a3'])[1:]
        S2=N.array(mydata1.data['a4'])[1:]
        A2=N.array(mydata1.data['a5'])[1:]*2
        M2=N.array(mydata1.data['a2'])[1:]
        Counts_mp=N.array(mydata1.data['detector'])[1:]
        timestamp_mp=N.array(mydata1.data['timestamp'])[1:]
        #H_mp,K_mp,L_mp,E_mp,Q_mp,Ei_mp,Ef_mp=mylattice.SpecWhere(myradians(M2),myradians(S1),myradians(S2),myradians(A2),setup)
        monlist.append(mydata1.data['monitor'][0])
        #print M2.shape
        #print Counts_mp.shape
        #remove potential bad datapoint that bt7 generated in dd mode


        myfilestr=flist2[0]
        mydatareader=readncnr.datareader()
        mydata2=mydatareader.readbuffer(myfilestr)
        S1=N.concatenate((S1,N.array(mydata2.data['a3'])[1:]))
        S2=N.concatenate((S2,N.array(mydata2.data['a4'])[1:]))
        A2=N.concatenate((A2,N.array(mydata2.data['a5'])[1:]*2))
        M2=N.concatenate((M2,N.array(mydata2.data['a2'])[1:]))
        timestamp_mp=N.concatenate((timestamp_mp,N.array(mydata2.data['timestamp'])[1:]))
        timestamp['mp']=timestamp_mp
        mylattice=lattice_calculator.lattice(a=a,b=b,c=c,alpha=alpha,beta=beta,gamma=gamma,\
                               orient1=orient1,orient2=orient2)

        Counts_mp=N.concatenate((Counts_mp,N.array(mydata2.data['detector'])[1:]))
        #print 'where'
        H_mp,K_mp,L_mp,E_mp,Q_mp,Ei_mp,Ef_mp=mylattice.SpecWhere(myradians(M2),myradians(S1),myradians(S2),myradians(A2),setup)


        ylist=[Counts_pm,Counts_mp]
        yerrlist=[N.sqrt(Counts_pm),N.sqrt(Counts_mp)]
        I,Ierr=simple_combine.monitor_normalize(ylist,yerrlist,monlist)
        counts={}
        errs={}
        counts['pm']=I[0]
        counts['mp']=I[1]
        errs['pm']=Ierr[0]
        errs['mp']=Ierr[1]
        cell=mydirectory+'\WilliamOct2007horizCells.txt'
        print cell
        pbflags=polcorrect_lib.PBflags()
        pbflags.MonitorCorrect=0
        pbflags.PolMonitorCorrect=1
        pbflags.MonoSelect=1
        pbflags.Debug=0
        pbflags.SimFlux=0
        pbflags.SimDeviate=0
        pbflags.NoNegativeCS=0
        pbflags.HalfPolarized=0
        pbflags.CountsEnable=[0,0,1,1]
        pbflags.CountsAdd1=[0,0,0,0]
        pbflags.CountsAdd2=[0,0,0,0]
        pbflags.Sconstrain=[1,1,0,0]
        pbflags.Spp=[0,0,0,0]
        pbflags.Smm=[0,0,0,0]
        pbflags.Spm=[0,0,0,0]
        pbflags.Smp=[0,0,0,0]

        mypolcor=polcorrect_lib.polarization_correct(counts,errs,timestamp,cell,Ei_mp,Ef_mp)
        corrected_counts=mypolcor.correct(pbflags)

        #print timestamp['mp'].shape
        #print timestamp['pm'].shape
        #print S1.shape
        #print S2.shape
        #print A2.shape
        #print M2.shape
        #print Counts_mp.shape
        #print H_mp.shape
        #print Counts_mp.shape
        #print I[0].shape
        #print I[1].shape
        if 1:
            outputfile=r'c:\bifeo3xtal\hhl_pm.txt'
            f=open(outputfile,'wt')
            f.write('H K L I Ierr\n')
            for i in range(H_pm.size):
                s='%f %f %f %f %f'%(H_pm[i],K_pm[i],L_pm[i],corrected_counts['Spm'][i],corrected_counts['Epm'][i])
                print s
                f.write(s+'\n')
            f.close()
        if 1:
            outputfile=r'c:\bifeo3xtal\hhl_mp.txt'
            f=open(outputfile,'wt')
            f.write('H K L I Ierr\n')
            for i in range(H_mp.size):
                s='%f %f %f %f %f'%(H_mp[i],K_mp[i],L_mp[i],corrected_counts['Smp'][i],corrected_counts['Emp'][i])
                print s
                f.write(s+'\n')
            f.close()

# Next zone
        myend='out'
        myend='bt7'
        mydirectory=r'c:\bifeo3xtal\jan8_2008\9175\data'
        myfilebase='fieldscansplusminusreset53630'
        myfilebase2='fieldscanminusplusreset53631'

        myfilebaseglob=myfilebase+'*.'+myend
        flist = SU.ffind(mydirectory, shellglobs=(myfilebaseglob,))
        myfilebaseglob=myfilebase2+'*.'+myend
        flist2 = SU.ffind(mydirectory, shellglobs=(myfilebaseglob,))

        myfilestr=flist[0]
        mydatareader=readncnr.datareader()
        mydata_pm=mydatareader.readbuffer(myfilestr)

        q_pm=N.array(mydata_pm.data['qx'])
        qh_pm=N.array(mydata_pm.data['qx'])
        qk_pm=N.array(mydata_pm.data['qy'])
        ql_pm=N.array(mydata_pm.data['qz'])


        a=N.array([3.97296],'d')
        b=N.array([2.79045],'d')
        c=N.array([13.86],'d')
        alpha=N.array([90],'d')
        beta=N.array([90],'d')
        gamma=N.array([90],'d')
        orient1=N.array([[1,0,0]],'d')
        orient2=N.array([[0,1,0]],'d')


        mylattice=lattice_calculator.lattice(a=a,b=b,c=c,alpha=alpha,beta=beta,gamma=gamma,\
                               orient1=orient1,orient2=orient2)
        Q_pm=mylattice.modvec(qh_pm,qk_pm,ql_pm,'latticestar')
        #counts_pm=N.array(mydata_pm.data['detector_corrected'])
        #errs_pm=N.array(mydata_pm.data['detector_errs_corrected'])
        counts_pm=N.array(mydata_pm.data['detector'])
        errs_pm=N.sqrt(counts_pm)


        ht_pm=-qh_pm+qk_pm
        kt_pm=qh_pm+qk_pm
        lt_pm=-2*qh_pm
        qh_pm=ht_pm
        qk_pm=kt_pm
        ql_pm=lt_pm
        q_pm=qh_pm




        myfilestr=flist2[0]
        mydatareader=readncnr.datareader()
        mydata_mp=mydatareader.readbuffer(myfilestr)
        q_mp=N.array(mydata_mp.data['qx'])
        qh_mp=N.array(mydata_mp.data['qx'])
        qk_mp=N.array(mydata_mp.data['qy'])
        ql_mp=N.array(mydata_mp.data['qz'])



        mylattice=lattice_calculator.lattice(a=a,b=b,c=c,alpha=alpha,beta=beta,gamma=gamma,\
                               orient1=orient1,orient2=orient2)
        Q_mp=mylattice.modvec(qh_mp,qk_mp,ql_mp,'latticestar')

        #counts_mp=N.array(mydata_mp.data['detector_corrected'])
        #errs_mp=N.array(mydata_mp.data['detector_errs_corrected'])

        counts_mp=N.array(mydata_mp.data['detector'])
        errs_mp=N.sqrt(counts_mp)

        ht_mp=-qh_mp+qk_mp
        kt_mp=qh_mp+qk_mp
        lt_mp=-2*qh_mp
        qh_mp=ht_mp
        qk_mp=kt_mp
        ql_mp=lt_mp
        q_mp=qh_mp



        cartoon()

        timestamp={}
        errs={}
        counts={}
        timestamp['mp']=N.array(mydata_mp.data['timestamp'])
        timestamp['pm']=N.array(mydata_pm.data['timestamp'])
        counts['pm']=counts_pm
        counts['mp']=counts_mp
        errs['pm']=errs_pm
        errs['mp']=errs_mp
        Ei_mp=14.7*N.ones(qh_pm.shape,'float64')
        Ef_mp=14.7*N.ones(qh_pm.shape,'float64')

        cell=r'C:\mytripleaxisproject\polarization\cells.txt'
        mypolcor=polcorrect_lib.polarization_correct(counts,errs,timestamp,cell,Ei_mp,Ef_mp)
        corrected_counts2=mypolcor.correct(pbflags)

        if 1:
            outputfile=r'c:\bifeo3xtal\hkhminusk_pm.txt'
            f=open(outputfile,'wt')
            f.write('Q H K L I Ierr\n')
            for i in range(qh_pm.size):
                #s='%f %f %f %f %f'%(qh_pm[i],qk_pm[i],ql_pm[i],counts_pm[i],errs_pm[i])
                s='%f %f %f %f %f %f'%(Q_pm[i],qh_pm[i],qk_pm[i],ql_pm[i],corrected_counts2['Spm'][i],corrected_counts2['Epm'][i])
                f.write(s+'\n')
                print s
            f.close()

        if 1:
            outputfile=r'c:\bifeo3xtal\hkhminusk_mp.txt'
            f=open(outputfile,'wt')
            f.write('Q H K L I Ierr\n')

            for i in range(qh_mp.size):
                #s='%f %f %f %f %f'%(qh_mp[i],qk_mp[i],ql_mp[i],counts_mp[i],errs_mp[i])
                s='%f %f %f %f %f %f'%(Q_mp[i],qh_mp[i],qk_mp[i],ql_mp[i],corrected_counts2['Smp'][i],corrected_counts2['Emp'][i])
                print s
                f.write(s+'\n')
                print s
            f.close()

        #exit()
        #print qh_mp.shape

        if 1:
            pylab.subplot(3,2,6)
            #pylab.errorbar(q_mp,counts_mp,fmt='bo',yerr=errs_mp,linestyle='None')
            #pylab.errorbar(q_pm,counts_pm,fmt='ro',yerr=errs_pm,linestyle='None')
            pylab.errorbar(q_mp,corrected_counts2['Smp'],fmt='bo',yerr=corrected_counts2['Emp'],linestyle='None')
            pylab.errorbar(q_pm,corrected_counts2['Spm'],fmt='ro',yerr=corrected_counts2['Epm'],linestyle='None')


        if 1:
            pylab.subplot(3,2,5)
            pylab.errorbar(H_mp,corrected_counts['Smp'],fmt='bo',yerr=corrected_counts['Emp'],linestyle='None')
            pylab.errorbar(H_pm,corrected_counts['Spm'],fmt='ro',yerr=corrected_counts['Epm'],linestyle='None')

        if 0:
            pylab.errorbar(H_mp,data['mp'],fmt='bo',yerr=errs['mp'],linestyle='None')
            pylab.errorbar(H_pm,data['pm'],fmt='ro',yerr=errs['pm'],linestyle='None')
            pylab.show()

        strangezone(fig)
        pylab.subplots_adjust(wspace=0.5,hspace=0.6)
        pylab.show()
        exit()
        if 0:
            for i in range(S1.size):
                s='%f %f %f %f'%(S1[i],S2[i],A2[i],M2[i])
                print s
            exit()
        if 0:
            H,K,L,E,Q,Ei,Ef=mylattice.SpecWhere(myradians([M2[0]]),myradians([-163.193]),myradians([29.833]),myradians([A2[0]]),setup)
            s='%f %f %f %f %f %f %f'%(S1[0],S2[0],H[0],K[0],L[0],E[0],Q[0])
            print s
        if 0:
            H,K,L,E,Q,Ei,Ef=mylattice.SpecWhere(myradians(M2),myradians(S1),myradians(S2),myradians(A2),setup)
            for i in range(1,S1.size):
                s='%f %f %f %f %f %f %f'%(S1[i],S2[i],H[i],K[i],L[i],E[i],Q[i])
                print s

        if 0:
            pylab.plot(H,L,'bo')
            pylab.xlabel('[1,1,0]')
            pylab.ylabel('[0,0,1]')
            #pylab.axis([-1,1,-4,4])
            pylab.show()




import numpy as N


def CleanArgs(**args):
    """Takes as input a set of arrays and/or scalars.  It returns a set of the arrays that are the length as the longest entry
    For shorter arrays, it pads them with the last value in the array.  It returns a dictionary based on the calling args"""
    npts=[]
    for name,val in args.iteritems():
        if type(val) in [type(13),type(13.0)]:
            args[name]=N.array([val],'float64')
        if type(val)==type([1,2,3]):
            npts.append(len(val))
        else:
            npts.append(args[name].shape[0])
    maxpts=max(npts)
    for name,val in args.iteritems():
        if type(val)==type([1,2,3]):
            for i in range(maxpts-len(val)):
                args[name].append(val[-1])
        else:
            if val.shape[0]<maxpts:
                last_val=val[-1]
                if len(val.shape)==1:
                    addendum=last_val*N.ones((maxpts-val.shape[0],),'float64')
                    args[name]=N.concatenate((val,addendum))
                elif len(val.shape)==2:
                    addendum=N.tile(last_val,(maxpts-val.shape[0],1))
                    args[name]=N.concatenate((val,addendum))
            #print name, len(val.shape),args[name]
    return args


if __name__=='__main__':
    a=N.array([1])
    b=N.array([1,2])
    c=90
    orient1=N.array([[1,2,3]])
    d=[1]
    myinput={'a':a,'b':b,'c':c}
    myoutput=CleanArgs(a=a,b=b,c=c,orient1=orient1,d=d)
    print a,b,c,orient1
    print myoutput['d']
    #print a2,b2,c2





from numpy import *

def SqwDemo(H,K,L,W,p):
    """This is an example of a cross section function for use with ConvRes.m.
    This particular function calculates the cross section for a gapped
    excitations in a 1-dimensional antiferromagnet. "a" is the chain axis.
    The polarization factors for each mode are NOT calculated here, but
    should be included in the prefactor function instead. This function is
    meant to be used together with the prefactor function PrefDemo.m
    Arguments H-W and are vectors, so dont forget to use ".*" instead of "*", etc."""


    #% Extract the three parameters contained in "p":
    Deltax=p[0] #					% Gap at the AF zone-center in meV for x-axis mode
    Deltay=p[1]	#				% Gap at the AF zone-center in meV for y-axis mode
    Deltaz=p[2]	#				% Gap at the AF zone-center in meV for z-axis mode
    cc=p[3] #						% Bandwidth in meV
    Gamma=p[4] #					    % Intrinsic HWHM of exccitation in meV
    #I=p[6);							% Intensity prefactor It will be used in the PrefDemo function, not here!
    #bgr=p[7);						% Background. It will be used in the PrefDemo function, not here!

    #% Calculate excitation energy at the (H,K,L) position for each of the three modes:
    omegax=sqrt(cc**2*(sin(2*pi*H))**2+Deltax**2)
    omegay=sqrt(cc**2*(sin(2*pi*H))**2+Deltay**2)
    omegaz=sqrt(cc**2*(sin(2*pi*H))**2+Deltaz**2)

    #% Assume Lorenzian excitation broadening:
    lorx=1./pi*Gamma/( (W-omegax)**2+Gamma**2 )
    lory=1./pi*Gamma/( (W-omegay)**2+Gamma**2 )
    lorz=1./pi*Gamma/( (W-omegaz)**2+Gamma**2 )


    # Intensity scales as (1-cos(2*pi*H))/omega0 for each of the three modes:
    sqw=zeros((1,size(H)),'float64')
    #print 'sqw ',sqw.shape
    #print 'lorx ',lorx.shape
    #print 'done'
    sqw[0,:]=lorx*(1-cos(pi*H))/omegax/2
    #sqw[1,:]=lory*(1-cos(pi*H))/omegay/2
    #sqw[2,:]=lorz*(1-cos(pi*H))/omegaz/2
    #print 'sqw ',sqw
    return sqw



import numpy as N
import math
import pylab
import unittest
eps=1e-3
pi=N.pi

def autovectorized(f):
     """Function decorator to do vectorization only as necessary.
     vectorized functions fail for scalar inputs."""
     def wrapper(input):
         if N.isscalar(input)==False:
             return N.vectorize(f)(input)
         return f(input)
     return wrapper



@autovectorized
def myradians(x):
    return math.radians(x)

#vecradians = N.vectorize(myradians, otypes=[double]) 

def sign(x):
    if x>0:
         ret=1
    if x<0:
        ret=-1
    if x==0:
        ret=0
    return ret

def blkdiag(g):
    "Returns a block diagonal matrix, given a list of square matrices, g"
    glen=len(g)
    n=0
    for i in range(glen):
   #     print g[i]
        n=n+g[i].shape[0]
   #     print n


    gout=N.zeros((n,n))
    offset=0
    for i in range(glen):
        currblock=g[i]
        lenx,leny=currblock.shape
    #    print i
    #    print lenx, leny
        for x in range(lenx):
            for y in range(leny):
                gout[x+offset,y+offset]=currblock[x,y]
     #           print gout
        offset=offset+lenx
    return gout


def similarity_transform(A,B):
    G=N.dot(B,A.transpose())
    G2=N.dot(A,G)
    return G2

class instrument:
    def __init__(self):
            self.tau_list={'pg(002)':1.87325, \
                            'pg(004)':3.74650, \
                            'ge(111)':1.92366, \
                            'ge(220)':3.14131, \
                            'ge(311)':3.68351, \
                            'be(002)':3.50702, \
                            'pg(110)':5.49806}
            #self.tau=self.tau_list[tau]
            return
    def get_tau(self,tau):
        return self.tau_list[tau]

class lattice:
    def __init__(self, a=N.array([2*N.pi, 2*N.pi], 'd'), \
                 b=N.array([2*N.pi, 2*N.pi], 'd'), \
                 c=N.array([2*N.pi, 2*N.pi], 'd'), \
                 alpha=N.array([90, 90], 'd'), \
                 beta=N.array([90, 90], 'd'), \
                 gamma=N.array([90, 90], 'd'), \
                 orient1=N.array([[1, 0, 0], [1, 0, 0]], 'd'), \
                 orient2=N.array([[0, 1, 0], [0, 1, 0]], 'd')):
        self.a=a
        self.b=b
        self.c=c
        self.alphad=alpha
        self.betad=beta
        self.gammad=gamma
        self.alpha=myradians(alpha)
        self.beta=myradians(beta)
        self.gamma=myradians(gamma)
        self.star()
        self.gtensor('lattice')
        self.gtensor('latticestar')
        self.npts=N.size(a)
        self.orient1=orient1.transpose()
        self.orient2=orient2.transpose()
        self.StandardSystem()
        return

    def scalar(self, x1, y1, z1, x2, y2, z2, lattice):
        "calculates scalar product of two vectors"
        if lattice=='lattice':
            a=self.a
            b=self.b
            c=self.c
            alpha=self.alpha
            beta=self.beta
            gamma=self.gamma
        if lattice=='latticestar':
            a=self.astar
            b=self.bstar
            c=self.cstar
            alpha=self.alphastar
            beta=self.betastar
            gamma=self.gammastar
        
        s=x1*x2*a**2+y1*y2*b**2+z1*z2*c**2+\
           (x1*y2+x2*y1)*a*b*N.cos(gamma)+\
           (x1*z2+x2*z1)*a*c*N.cos(beta)+\
           (z1*y2+z2*y1)*c*b*N.cos(alpha)
        return s

    def star(self):
        "Calculate unit cell volume, reciprocal cell volume, reciprocal lattice parameters"
        V=2*self.a*self.b*self.c*\
        N.sqrt(N.sin((self.alpha+self.beta+self.gamma)/2)*\
                N.sin((-self.alpha+self.beta+self.gamma)/2)*\
                N.sin((self.alpha-self.beta+self.gamma)/2)*\
                N.sin((self.alpha+self.beta-self.gamma)/2))
        self.Vstar=(2*N.pi)**3/V;
        self.astar=2*N.pi*self.b*self.c*N.sin(self.alpha)/V
        self.bstar=2*N.pi*self.a*self.c*N.sin(self.beta)/V
        self.cstar=2*N.pi*self.b*self.a*N.sin(self.gamma)/V
        self.alphastar=N.arccos((N.cos(self.beta)*N.cos(self.gamma)-\
                                 N.cos(self.alpha))/ \
                                (N.sin(self.beta)*N.sin(self.gamma)))
        self.betastar= N.arccos((N.cos(self.alpha)*N.cos(self.gamma)-\
                                 N.cos(self.beta))/ \
                                (N.sin(self.alpha)*N.sin(self.gamma)))
        self.gammastar=N.arccos((N.cos(self.alpha)*N.cos(self.beta)-\
                               N.cos(self.gamma))/ \
                              (N.sin(self.alpha)*N.sin(self.beta)))
        self.V=V
        return

    def angle2(self, x, y, z, h, k, l):
        "Calculate the angle between vectors in real and reciprocal space"
        "x,y,z are the fractional cell coordinates of the first vector,"
        "h,k,l are Miller indices of the second vector"
        phi=N.arccos(2*pi*(h*x+k*y+l*z)/self.modvec(x, y, z, 'lattice')/self.modvec(h, k, l, 'latticestar'))
        return phi


    def angle(self, x1, y1, z1, x2, y2, z2, lattice):
        "Calculate the angle between vectors in real and reciprocal space"
        "xi,yi,zi are the fractional cell coordinates of the vectors"
        phi=N.arccos(self.scalar(x1, y1, z1, x2, y2, z2, lattice)/self.modvec(x1, y1, z1, lattice)/self.modvec(x1, y1, z1, lattice))    
        return phi
    
    def modvec(self, x, y, z, lattice):
        "Calculates modulus of a vector defined by its fraction cell coordinates"
        "or Miller indexes"
        m=N.sqrt(self.scalar(x, y, z, x, y, z, lattice))
        return m

    def gtensor(self, lattice):
        "calculates the metric tensor of a lattice"
        g=N.zeros((3, 3, N.size(self.a)), 'd')
        #print 'shape ', g.shape
        if lattice=='lattice':
            a=self.a
            b=self.b
            c=self.c
            alpha=self.alpha
            beta=self.beta
            gamma=self.gamma
        if lattice=='latticestar':
            a=self.astar
            b=self.bstar
            c=self.cstar
            alpha=self.alphastar
            beta=self.betastar
            gamma=self.gammastar
        g[0, 0, :]=a**2;
        g[0, 1, :]=a*b*N.cos(gamma)
        g[0, 2, :]=a*c*N.cos(beta)

        g[1, 0, :]=g[0, 1, :]
        g[1, 1, :]=b**2
        g[1, 2, :]=c*b*N.cos(alpha)

        g[2, 0, :]=g[0, 2, :]
        g[2, 1, :]=g[1, 2, :]
        g[2, 2, :]=c**2
        if lattice=='lattice':
            self.g=g
#            print 'lattice'
        if lattice=='latticestar':
            self.gstar=g
#            print 'latticestar'
#        print g
        return

    def reciprocate(self, x, y, z, lattice):
        "calculate miller indexes of a vector defined by its fractional cell coords"
        if lattice=='lattice':
            g=self.g
        if lattice=='latticestar':
            g=self.gstar
        h=g[0, 0, :]*x+g[1, 0, :]*y+g[2, 0, :]*z;
        k=g[0, 1, :]*x+g[1, 1, :]*y+g[2, 1, :]*z;
        l=g[0, 2, :]*x+g[1, 2, :]*y+g[2, 2, :]*z;
        return h, k, l

    def vector(self, x1, y1, z1, x2, y2, z2, lattice):
        "calculates the fractional cell coordinates or Miller indexes of a vector"
        "product of two vectors, defined by their fractional cell coordinates or "
        "Miller idexes"
        if lattice=='lattice':
            g=self.gstar
            V=self.Vstar
        if lattice=='latticestar':
            g=self.g
            V=self.V
        g=g*V/(2*N.pi)**2
        x=y1*z2*g[0, 0, :]-z1*y2*g[0, 0, :]-x1*z2*g[1, 0, :]+z1*x2*g[1, 0, :]\
           +x1*y2*g[2, 0, :]-y1*x2*g[2, 0, :]
        y=y1*z2*g[0, 1, :]-z1*y2*g[0, 1, :]-x1*z2*g[1, 1, :]+z1*x2*g[1, 1, :]\
           +x1*y2*g[2, 1, :]-y1*x2*g[2, 1, :]
        z=y1*z2*g[0, 2, :]-z1*y2*g[0, 2, :]-x1*z2*g[1, 2, :]+z1*x2*g[1, 2, :]\
           +x1*y2*g[2, 2, :]-y1*x2*g[2, 2, :]
        return x,y,z

    def StandardSystem(self):
#        npts=self.npts
#        orient1=N.zeros((3,npts),'d')
#        orient2=N.zeros((3,npts),'d')
        orient1=self.orient1
        orient2=self.orient2
#        print 'orient1'
#        print orient1
#        print 'orient2 '
#        print orient2        
        modx=self.modvec(orient1[0, :], orient1[1, :], orient1[2, :], 'latticestar')
        x=N.copy(orient1)
        x[0, :]=x[0, :]/modx; # First unit basis vector
        x[1, :]=x[1, :]/modx;
        x[2, :]=x[2, :]/modx;
        
        proj=self.scalar(orient2[0, :], orient2[1, :], orient2[2, :], \
                    x[0, :], x[1, :], x[2, :], 'latticestar')
        
        y=N.copy(orient2)
        y[0, :]=y[0, :]-x[0, :]*proj; 
        y[1, :]=y[1, :]-x[1, :]*proj;
        y[2, :]=y[2, :]-x[2, :]*proj;

        mody=self.modvec(y[0, :], y[1, :], y[2, :], 'latticestar');

#    check for collinearity of orienting vectors
# implement later!!!
#        if ~isempty(find(mody<=0))
#        error('??? Fatal error: Orienting vectors are colinear!');
#        print 'where '+str((N.where(mody<=0)[0].size))
#        print 'mody '
#        print  mody
        try:
#            print N.where(mody<=eps)[0].size
            if N.where(mody<=eps)[0].size>0:
                print 'ValueError'
                raise ValueError
            y[0, :]=y[0, :]/mody; # Second unit basis vector
            y[1, :]=y[1, :]/mody;
            y[2, :]=y[2, :]/mody;
    
            z=N.copy(y);
    
            z[0, :]=x[1, :]*y[2, :]-y[1, :]*x[2, :];
            z[1, :]=x[2, :]*y[0, :]-y[2, :]*x[0, :];
            z[2, :]=-x[1, :]*y[0, :]+y[1, :]*x[0, :];
    
            proj=self.scalar(z[0, :], z[1, :], z[2, :], x[0, :], x[1, :], x[2, :], 'latticestar');
    
            z[0, :]=z[0, :]-x[0, :]*proj; 
            z[1, :]=z[1, :]-x[1, :]*proj;
            z[2, :]=z[2, :]-x[2, :]*proj;
    
            proj=self.scalar(z[0, :], z[1, :], z[2, :], y[0, :], y[1, :], y[2, :], 'latticestar');
    
            z[0, :]=z[0, :]-y[0, :]*proj; 
            z[1, :]=z[1, :]-y[1, :]*proj;
            z[2, :]=z[2, :]-y[2, :]*proj;
    
            modz=self.modvec(z[0, :], z[1, :], z[2, :], 'latticestar');
    
            z[0, :]=z[0, :]/modz; #% Third unit basis vector
            z[1, :]=z[1, :]/modz;
            z[2, :]=z[2, :]/modz;     
            
            self.x=x
            self.y=y
            self.z=z    
        except ValueError:
            print 'ORIENTATION VECTORS ARE COLLINEAR x,y,z not set'    
        return
    
    def S2R(self, qx, qy, qz):
        "Given cartesian coordinates of a vector in the S System, calculate its Miller indexes."
        x=self.x
        y=self.y
        z=self.z
        H=qx*x[0, :]+qy*y[0, :]+qz*z[0, :];
        K=qx*x[1, :]+qy*y[1, :]+qz*z[1, :];
        L=qx*x[2, :]+qy*y[2, :]+qz*z[2, :];
        q=N.sqrt(qx**2+qy**2+qz**2);
        return H, K, L, q
    
    def R2S(self, H, K, L):
        "Given reciprocal-space coordinates of a vecotre, calculate its coordinates in the Cartesian space."
        x=self.x
        y=self.y
        z=self.z
        qx=self.scalar(H, K, L, x[0, :], x[1, :], x[2, :], 'latticestar');
        qy=self.scalar(H, K, L, y[0, :], y[1, :], y[2, :], 'latticestar');
        qz=self.scalar(H, K, L, z[0, :], z[1, :], z[2, :], 'latticestar');
        q=self.modvec(H, K, L, 'latticestar');
        return qx, qy, qz, q
    
    def ResMat(self, Q, W,EXP):
        CONVERT1=0.4246609*N.pi/60/180;
        CONVERT2=2.072;
        npts=self.npts
        RM=N.zeros((4, 4, npts),'d');
        R0=N.zeros((1, npts),'d');
        RM_=N.zeros((4, 4),'d');
        D=N.zeros((8, 13),'d');
        d=N.zeros((4, 7),'d');
        T=N.zeros((4, 13),'d');
        t=N.zeros((2, 7),'d');
        A=N.zeros((6, 8),'d');
        C=N.zeros((4, 8),'d');
        B=N.zeros((4, 6),'d');

        for ind in range(npts):
#            %Assign default values and decode parameters
            moncor=1;
            if 'moncor' in EXP[ind]:
                moncor = EXP[ind]['moncor']
            alpha = EXP[ind]['hcol']*CONVERT1;
            beta =  EXP[ind]['vcol']*CONVERT1;
            mono=EXP[ind]['mono']
            etam = mono['mosaic']*CONVERT1;
            etamv=etam
            if 'vmosaic' in mono:
                etamv = mono['vmosaic']*CONVERT1;
            ana=EXP[ind]['ana'];
            etaa = ana['mosaic']*CONVERT1;
            etaav=etaa;
            if 'vmosaic' in ana:
                etaav = ana['vmosaic']*CONVERT1;
            sample=EXP[ind]['sample'];
            infin=-1;
            if 'infin' in EXP[ind]:
                infin = EXP[ind]['infin']
            efixed=EXP[ind]['efixed']
            epm=1
            if 'dir1' in EXP[ind]:
                epm= EXP[ind]['dir1'];
            ep=1;
            if 'dir2' in EXP[ind]:
                ep= EXP[ind]['dir2'];
            monitorw=1;
            monitorh=1;
            beamw=1;
            beamh=1;
            monow=1;
            monoh=1;
            monod=1;
            anaw=1;
            anah=1;
            anad=1;
            detectorw=1;
            detectorh=1;
            sshape=N.eye(3);
            L0=1;
            L1=1;
            L1mon=1;
            L2=1;
            L3=1;        
            monorv=1e6;
            monorh=1e6;
            anarv=1e6;
            anarh=1e6;
            if 'beam' in EXP[ind]:
                beam=EXP[ind]['beam'];
                if 'width' in beam:
                    beamw=beam['width']**2;
                if 'height' in beam:
                    beamh=beam['height']**2
            bshape=N.diag([beamw,beamh]);
            if 'monitor' in EXP[ind]:
                monitor=EXP[ind]['monitor'];
                if 'width' in monitor:
                    monitorw=monitor['width']**2
                monitorh=monitorw;
                if 'height' in monitor:
                    monitorh=monitor['height']**2;
            monitorshape=N.diag([monitorw,monitorh]);
            if 'detector' in EXP[ind]:
                detector=EXP[ind]['detector'];
                if 'width' in detector:
                    detectorw=detector['width']**2;
                if 'height' in detector:
                    detectorh=detector['height']**2;
            dshape=N.diag([detectorw,detectorh]);
            if 'width' in mono:
                monow=mono['width']**2;
            if 'height' in mono:
                monoh=mono['height']**2;
            if 'depth' in mono:
                monod=mono.depth**2;
            mshape=N.diag([monod,monow,monoh]);
            if 'width' in ana: 
                anaw=ana['width']**2;
            if 'height' in ana:
                anah=ana['height']**2;
            if 'depth' in ana:
                anad=ana['depth']**2;
            ashape=N.diag([anad,anaw,anah]);
            if 'shape' in sample:
                sshape=sample['shape'];
            if 'arms' in EXP[ind]:
                arms=EXP[ind]['arms'];
                L0=arms[0];
                L1=arms[2];
                L2=arms[3];
                L3=arms[4];
                L1mon=L1;
                if len(arms)>3:
                    L1mon=arms[4];
            if 'rv' in mono:
                monorv=mono['rv'];
            if 'rh' in mono:
                monorh=mono['rh'];
            if 'rv' in ana:
                anarv=ana['rv'];
            if 'rh' in ana:
                anarh=ana['rh'];
            method=0;
            if 'method' in EXP[ind]:
                method=EXP[ind]['method'];
            myinstrument=instrument()
            taum=myinstrument.get_tau(mono['tau']);
            taua=myinstrument.get_tau(ana['tau']);
        
            horifoc=-1;
            if 'horifoc' in EXP[ind]:
                horifoc=EXP[ind]['horifoc'];
            if horifoc==1: 
                alpha[2]=alpha[2]*N.sqrt(8*N.log(2)/12); 
#            %---------------------------------------------------------------------------------------------
#            %Calculate angles and energies
            w=W[ind];
            q=Q[ind];
            ei=efixed;
            ef=efixed;
            if infin>0:
                 ef=efixed-w
            else:
                 ei=efixed+w; 
            ki = N.sqrt(ei/CONVERT2);
            kf = N.sqrt(ef/CONVERT2);
            thetam=N.arcsin(taum/(2*ki))*sign(epm); 
            thetaa=N.arcsin(taua/(2*kf))*sign(ep); 
            s2theta=-N.arccos( (ki**2+kf**2-q**2)/(2*ki*kf));# %2theta sample
            thetas=s2theta/2;
            phi=N.arctan2(-kf*N.sin(s2theta), ki-kf*N.cos(s2theta)); #%Angle from ki to Q
        
 #           %---------------------------------------------------------------------------------------------
#            %Calculate beam divergences defined by neutron guides
            pi=N.pi
            if alpha[0]<0:
                  alpha[0]=-alpha[0]*2*0.427/ki*pi/180; 
            if alpha[1]<0:
                  alpha[1]=-alpha[1]*2*0.427/ki*pi/180; 
            if alpha[2]<0:
                  alpha[2]=-alpha[2]*2*0.427/ki*pi/180; 
            if alpha[3]<0:
                  alpha[3]=-alpha[3]*2*0.427/ki*pi/180; 
            
            if beta[0]<0:
                  beta[0]=-beta[0]*2*0.427/ki*pi/180; 
            if beta[1]<0:
                  beta[1]=-beta[1]*2*0.427/ki*pi/180; 
            if beta[2]<0:
                  beta[2]=-beta[2]*2*0.427/ki*pi/180; 
            if beta[3]<0:
                  beta[3]=-beta[3]*2*0.427/ki*pi/180; 
            
#            %---------------------------------------------------------------------------------------------
#            %Rededine sample geometry
            psi=thetas-phi;# %Angle from sample geometry X axis to Q
            rot=N.zeros((3,3));
            rot[0,0]=N.cos(psi);
            rot[1,1]=N.cos(psi);
            rot[0,1]=N.sin(psi);
            rot[1,0]=-N.sin(psi);
            rot[2,2]=1;
            sshape=N.dot(rot.transpose(),N.dot(sshape,rot)); #matrix multiplication?   
#            %---------------------------------------------------------------------------------------------
#            %Definition of matrix G    
            G=1.0/N.array([alpha[0],alpha[1],beta[0],beta[1],alpha[2],alpha[3],beta[2],beta[3]])**2;
            G=N.diag(G);
#            %---------------------------------------------------------------------------------------------
#            %Definition of matrix F    
            F=1.0/N.array([etam,etamv,etaa,etaav])**2;
            F=N.diag(F);
#            %---------------------------------------------------------------------------------------------
#            %Definition of matrix A
            A[0,0]=ki/2/N.tan(thetam);
            A[0,1]=-A[0,0];
            A[3,4]=kf/2/N.tan(thetaa);
            A[3,5]=-A[3,4];
            A[1,1]=ki;
            A[2,3]=ki;
            A[4,4]=kf;
            A[5,6]=kf;
#            %---------------------------------------------------------------------------------------------
#            %Definition of matrix C
            C[0,0]=1.0/2;
            C[0,1]=1.0/2;
            C[2,4]=1.0/2;
            C[2,5]=1.0/2;
            C[1,2]=1.0/(2*N.sin(thetam));
            C[1,3]=-C[1,2];# %mistake in paper
            C[3,6]=1.0/(2*N.sin(thetaa));
            C[3,7]=-C[3,6];
#            %---------------------------------------------------------------------------------------------
#            %Definition of matrix B
            B[0,0]=N.cos(phi);
            B[0,1]=N.sin(phi);
            B[0,3]=-N.cos(phi-s2theta);
            B[0,4]=-N.sin(phi-s2theta);
            B[1,0]=-B[0,1];
            B[1,1]=B[0,0];
            B[1,3]=-B[0,4];
            B[1,4]=B[0,3];
            B[2,2]=1.0;
            B[2,5]=-1.0;
            B[3,0]=2*CONVERT2*ki;
            B[3,3]=-2*CONVERT2*kf;
#            %---------------------------------------------------------------------------------------------
#            %Definition of matrix S
            Sinv=blkdiag([bshape,mshape,sshape,ashape,dshape]);# %S-1 matrix        
            S=N.linalg.inv(Sinv);
#            %---------------------------------------------------------------------------------------------
#            %Definition of matrix T
            T[0,0]=-1./(2*L0); # %mistake in paper
            T[0,2]=N.cos(thetam)*(1./L1-1./L0)/2;
            T[0,3]=N.sin(thetam)*(1./L0+1./L1-2./(monorh*N.sin(thetam)))/2;
            T[0,5]=N.sin(thetas)/(2*L1);
            T[0,6]=N.cos(thetas)/(2*L1);
            T[1,1]=-1./(2*L0*N.sin(thetam));
            T[1,4]=(1./L0+1./L1-2*N.sin(thetam)/monorv)/(2*N.sin(thetam));
            T[1,7]=-1./(2*L1*N.sin(thetam));
            T[2,5]=N.sin(thetas)/(2*L2);
            T[2,6]=-N.cos(thetas)/(2*L2);
            T[2,8]=N.cos(thetaa)*(1./L3-1./L2)/2;
            T[2,9]=N.sin(thetaa)*(1./L2+1./L3-2/(anarh*N.sin(thetaa)))/2;
            T[2,11]=1./(2*L3);
            T[3,7]=-1./(2*L2*N.sin(thetaa));
            T[3,10]=(1./L2+1./L3-2*N.sin(thetaa)/anarv)/(2*N.sin(thetaa));
            T[3,12]=-1./(2*L3*N.sin(thetaa));
#            %---------------------------------------------------------------------------------------------
#            %Definition of matrix D
#            % Lots of index mistakes in paper for matix D
            D[0,0]=-1./L0;
            D[0,2]=-N.cos(thetam)/L0;
            D[0,3]=N.sin(thetam)/L0;
            D[2,1]=D[0,0];
            D[2,4]=-D[0,0];
            D[1,2]=N.cos(thetam)/L1;
            D[1,3]=N.sin(thetam)/L1;
            D[1,5]=N.sin(thetas)/L1;
            D[1,6]=N.cos(thetas)/L1;
            D[3,4]=-1./L1;
            D[3,7]=-D[3,4];
            D[4,5]=N.sin(thetas)/L2;
            D[4,6]=-N.cos(thetas)/L2;
            D[4,8]=-N.cos(thetaa)/L2;
            D[4,9]=N.sin(thetaa)/L2;
            D[6,7]=-1./L2;
            D[6,10]=-D[6,7];
            D[5,8]=N.cos(thetaa)/L3;
            D[5,9]=N.sin(thetaa)/L3;
            D[5,11]=1./L3;
            D[7,10]=-D[5,11];
            D[7,12]=D[5,11];
#            %---------------------------------------------------------------------------------------------
#            %Definition of resolution matrix M
            if method==1:
                Minv=N.dot(N.dot(B,A),N.dot(N.linalg.inv(D.linalg.inv(D*N.linalg.inv(S+T.transpose()*F*T)*D.transpose())+G),\
                N.dot(A.transpose(),B.transpose())))#; %Popovici
                #Minv=B*A*((D*(S+T'*F*T)^(-1)*D')^(-1)+G)^(-1)*A'*B'; %Popovici
            else:
                #%Horizontally focusing analyzer if needed
                #print 'intermediate C.T*F*C'
                #print N.dot(C.transpose(),N.dot(F,C))
                #print 'product'
                #print N.dot(F,C)
                HF_int=N.linalg.inv(G+N.dot(C.transpose(),N.dot(F,C)));
                HF=similarity_transform(A,HF_int)
                #print 'HF'
                #print HF
                if horifoc>0:
                    HF=N.linalg.inv(HF);
                    HF[4,4]=(1.0/(kf*alpha[2]))**2; 
                    HF[4,3]=0; 
                    HF[3,4]=0; 
                    HF[3,3]=(N.tan(thetaa)/(etaa*kf))**2;
                    HF=N.linalg.inv(HF);
                Minv=similarity_transform(B,HF)#; %Cooper-Nathans
            M=N.linalg.inv(Minv);
            #print 'A'
            #print A
            #print 'B'
            #print B
            #print 'C'
            #print C
            #print 'D'
            #print D
            #print 'T'
            #print T
            #print 'M'
            #print M
            #print 'Minv'
            #print Minv
            #print 'G'
            #print G
            #print 'F'
            #print F
            RM_[0,0]=M[0,0];
            RM_[1,0]=M[1,0];
            RM_[0,1]=M[0,1];
            RM_[1,1]=M[1,1];
            
            RM_[0,2]=M[0,3];
            RM_[2,0]=M[3,0];
            RM_[2,2]=M[3,3];
            RM_[2,1]=M[3,1];
            RM_[1,2]=M[1,3];
            
            RM_[0,3]=M[0,2];
            RM_[3,0]=M[2,0];
            RM_[3,3]=M[2,2];
            RM_[3,1]=M[2,1];
            RM_[1,3]=M[1,2];
 #           %---------------------------------------------------------------------------------------------
 #           %Calculation of prefactor, normalized to source
            #print 'RM_'
            #print RM_
            Rm=ki**3/N.tan(thetam); 
            Ra=kf**3/N.tan(thetaa);
            #print 'Rm'
            #print Rm
            #print 'Ra'
            #print Ra
            if method==1:
                R0_=Rm*Ra*(2*pi)**4/(64*pi**2*N.sin(thetam)*N.sin(thetaa))\
                *N.sqrt(N.linalg.det(F)/N.linalg.det\
                        (N.linalg.inv(N.dot(D,N.dot(N.linalg.inv(S+N.dot(T.T,N.dot(F,T))),D.T)))+G)); #%Popovici
            else:
                R0_=Rm*Ra*(2*pi)**4/(64*pi**2*N.sin(thetam)*N.sin(thetaa))\
                *N.sqrt( N.linalg.det(F)/N.linalg.det(G+N.dot(C.transpose(),N.dot(F,C)))); #%Cooper-Nathans
                #print 'RO_'
                #print R0_
#            %---------------------------------------------------------------------------------------------
#            %Normalization to flux on monitor
            if moncor==1:
                g=G[0:4][:,0:4];
                f=F[0:2][:,0:2];
                #print 'f'
                #print f
                #print 'g'
                #print g
                c=C[0:2][:,0:4];
                t[0,0]=-1./(2*L0);#  %mistake in paper
                t[0,2]=N.cos(thetam)*(1./L1mon-1./L0)/2;
                t[0,3]=N.sin(thetam)*(1./L0+1./L1mon-2./(monorh*N.sin(thetam)))/2;
                t[0,6]=1./(2*L1mon);
                t[1,1]=-1./(2*L0*N.sin(thetam));
                t[1,4]=(1./L0+1./L1mon-2*N.sin(thetam)/monorv)/(2*N.sin(thetam));
                sinv=blkdiag([bshape,mshape,monitorshape]);# %S-1 matrix        
                s=N.linalg.inv(sinv);
                d[0,0]=-1./L0;
                d[0,2]=-N.cos(thetam)/L0;
                d[0,3]=N.sin(thetam)/L0;
                d[2,1]=D[0,0];
                d[2,4]=-D[0,0];
                d[1,2]=N.cos(thetam)/L1mon;
                d[1,3]=N.sin(thetam)/L1mon;
                d[1,5]=0;
                d[1,6]=1./L1mon;
                d[3,4]=-1./L1mon;
                if method==1:
                    Rmon=Rm*(2*pi)**2/(8*pi*N.sin(thetam))*N.sqrt(N.linalg.det(f)/\
                                                                  N.linalg.det(N.linalg.inv(N.dot(d,N.dot(N.linalg.inv(s+t.transpose()*f*t),d.transpose())))+g)); #%Popovici
                else:
                    Rmon=Rm*(2*pi)**2/(8*pi*N.sin(thetam))*N.sqrt(N.linalg.det(f)/N.linalg.det(g+N.dot(c.transpose(),N.dot(f,c)))); #%Cooper-Nathans
                R0_=R0_/Rmon;
                R0_=R0_*ki; #%1/ki monitor efficiency
                #print 'R01', R0_
                #print 'Rmon', Rmon
                
#            %---------------------------------------------------------------------------------------------
#            %Transform prefactor to Chesser-Axe normalization
            R0_=R0_/(2*pi)**2*N.sqrt(N.linalg.det(RM_));
#            %---------------------------------------------------------------------------------------------
#            %Include kf/ki part of cross section
            R0_=R0_*kf/ki;
#            %---------------------------------------------------------------------------------------------
#            %Take care of sample mosaic if needed [S. A. Werner & R. Pynn, J. Appl. Phys. 42, 4736, (1971)]
            if 'mosaic' in sample:
                etas = sample['mosaic']*CONVERT1;
                etasv=etas;
                if 'vmosaic' in sample:
                    etasv = sample['vmosaic']*CONVERT1;
                R0_=R0_/N.sqrt((1.+(q*etas)**2*RM_[3,3])*(1.0+(q*etasv)**2*RM_[1,1]));
                Minv=N.linalg.inv(RM_)
                Minv[1,1]=Minv[1,1]+q**2*etas**2;
                Minv[3,3]=Minv[3,3]+q**2*etasv**2;
                RM_=N.linalg.inv(Minv);
#            %---------------------------------------------------------------------------------------------
#            %Take care of analyzer reflectivity if needed [I. Zaliznyak, BNL]
            if ('thickness' in ana) & ('Q' in ana):
                KQ = ana['Q'];
                KT = ana['thickness'];
                toa=(taua/2)/N.sqrt(kf**2-(taua/2)**2);
                smallest=alpha[3];
                if alpha[3]>alpha[2]:
                     smallest=alpha[2]
                Qdsint=KQ*toa;
                dth=(N.arange(201)/200)*N.sqrt(2*N.log(2))*smallest;
                wdth=N.exp(-dth**2/2./etaa**2);
                sdth=KT*Qdsint*wdth/etaa/N.sqrt(2.*pi);
                rdth=1./(1+1./sdth);
                reflec=rdth.sum()/wdth.sum();
                R0_=R0_*reflec;
#            %---------------------------------------------------------------------------------------------
            R0[ind]=R0_;
            RM[:,:,ind]=RM_[:,:];
        return R0, RM
    
    def ResMatS(self,H,K,L,W,EXP):
# [len,H,K,L,W,EXP]=CleanArgs(H,K,L,W,EXP);
        x=self.x
        y=self.y
        z=self.z
        Q=self.modvec(H,K,L,'latticestar')
        uq=N.zeros((3,self.npts),'d')
        uq[0,:]=H/Q;  #% Unit vector along Q
        uq[1,:]=K/Q;
        uq[2,:]=L/Q;        
        xq=self.scalar(x[0,:],x[1,:],x[2,:],uq[0,:],uq[1,:],uq[2,:],'latticestar');
        yq=self.scalar(y[0,:],y[1,:],y[2,:],uq[0,:],uq[1,:],uq[2,:],'latticestar');
        zq=0; # %scattering vector assumed to be in (self.orient1,self.orient2) plane;        
        tmat=N.zeros((4,4,self.npts)); #%Coordinate transformation matrix
        tmat[3,3,:]=1;
        tmat[2,2,:]=1;
        tmat[0,0,:]=xq;
        tmat[0,1,:]=yq;
        tmat[1,1,:]=xq;
        tmat[1,0,:]=-yq;
        
        RMS=N.zeros((4,4,self.npts));
        rot=N.zeros((3,3));
        EXProt=EXP;
        
#        %Sample shape matrix in coordinate system defined by scattering vector
        for i in range(self.npts):
            sample=EXP[i]['sample'];
            if 'shape' in sample:
                rot[0,0]=tmat[0,0,i];
                rot[1,0]=tmat[1,0,i];
                rot[0,1]=tmat[0,1,i];
                rot[1,1]=tmat[1,1,i];
                rot[2,2]=tmat[2,2,i];
                EXProt[i]['sample']['shape']=N.dot(rot,N.dot(sample['shape'],rot.T));
        
        R0,RM= self.ResMat(Q,W,EXProt)
        
        for i in range(self.npts):
           RMS[:,:,i]=N.dot((tmat[:,:,i]).transpose(),N.dot(RM[:,:,i],tmat[:,:,i]));
        
        mul=N.zeros((4,4));
        e=N.eye(4,4);
        for i in range(self.npts):
            if 'Smooth' in EXP[i]:
                if 'X' in (EXP[i]['Smooth']):
                    mul[0,0]=1./(EXP[i]['Smooth']['X']**2/8/N.log(2));
                    mul[1,1]=1./(EXP[i]['Smooth']['Y']**2/8/N.log(2));
                    mul[2,2]=1./(EXP[i]['Smooth']['E']**2/8/N.log(2));
                    mul[3,3]=1./(EXP[i]['Smooth']['Z']**2/8/N.log(2));
                    R0[i]=R0[i]/N.sqrt(N.linalg.det(e/RMS[:,:,i]))*N.sqrt(N.linalg.det(e/mul+e/RMS[:,:,i]));
                    RMS[:,:,i]=e/(e/mul+e/RMS[:,:,i]);
        return R0, RMS
       
class TestLattice(unittest.TestCase):

    def setUp(self):
        a=N.array([2*pi],'d')
        b=N.array([2*pi],'d')
        c=N.array([2*pi],'d')
        alpha=N.array([90],'d')
        beta=N.array([90],'d')
        gamma=N.array([90],'d')
        orient1=N.array([[1,0,0]],'d')
        orient2=N.array([[0,1,1]],'d')
        self.fixture = lattice(a=a,b=b,c=c,alpha=alpha,beta=beta,gamma=gamma,\
                               orient1=orient1,orient2=orient2)
    
    def test_astar(self):
        self.assertAlmostEqual(self.fixture.astar[0],1.0,2,'astar Not equal to '+str(1.0))
    def test_bstar(self):
        self.assertAlmostEqual(self.fixture.bstar[0],1.0,2,'bstar Not equal to '+str(1.0))
    def test_cstar(self):
        self.assertAlmostEqual(self.fixture.cstar[0],1.0,2,'cstar '+str(self.fixture.cstar[0])+' Not equal to '+str(1.0))
    def test_alphastar(self):
        self.assertAlmostEqual(self.fixture.alphastar[0],pi/2,2,'alphastar Not equal to '+str(pi/2))
    def test_betastar(self):
        self.assertAlmostEqual(self.fixture.betastar[0],pi/2,2,'betastar Not equal to '+str(pi/2))
    def test_gammastar(self):
        self.assertAlmostEqual(self.fixture.gammastar[0],pi/2,2,'gammastar Not equal to '+str(pi/2))
    def test_V(self):
        self.assertAlmostEqual(self.fixture.V[0],248.0502,2,'V Not equal to '+str(248.0502))
    def test_Vstar(self):
        self.assertAlmostEqual(self.fixture.Vstar[0],1.0,2,'Vstar Not equal to '+str(1.0))
    def test_g(self):
        #print self.fixture.g
        self.assertAlmostEqual((self.fixture.g[:,:,0][0,0]),39.4784*(N.eye(3)[0,0]) ,2,'g Not equal to '+str(39.4784 ))
    def test_gstar(self):
        #print self.fixture.gstar
        self.assertAlmostEqual(self.fixture.gstar[:,:,0][0,0],1.0*N.eye(3)[0,0] ,2,'gstar Not equal to '+str(1.0 ))
 
    def test_StandardSystem_x(self):
 #       #print self.fixture.gstar
        self.assertAlmostEqual(self.fixture.x[0],1.0 ,2,'Standard System x Not equal to '+str(1.0 ))
 
    
      
                               
#    def test_zeroes(self):
#        self.assertEqual(0 + 0, 0)
#        self.assertEqual(5 + 0, 5)
#        self.assertEqual(0 + 13.2, 13.2)
#
#    def test_positive(self):
#        self.assertEqual(123 + 456, 579)
#        self.assertEqual(1.2e20 + 3.4e20, 3.5e20)
#
#    def test_mixed(self):
#        self.assertEqual(-19 + 20, 1)
#        self.assertEqual(999 + -1, 998)
#        self.assertEqual(-300.1 + -400.2, -700.3)
#        

#if __name__=="__main__":
    if 1:
        a=N.array([2*pi],'d')
        b=N.array([2*pi],'d')
        c=N.array([2*pi],'d')
        alpha=N.array([90],'d')
        beta=N.array([90],'d')
        gamma=N.array([90],'d')
 #       orient1=N.array([[0,1,1]],'d')
        orient1=N.array([[1,0,0]],'d')
        orient2=N.array([[0,1,1]],'d')
        mylattice=lattice(a=a,b=b,c=c,alpha=alpha,beta=beta,gamma=gamma,\
                               orient1=orient1,orient2=orient2)
        H=N.array([1],'d');K=N.array([0],'d');L=N.array([0],'d');W=N.array([0],'d')
        EXP={}
        EXP['ana']={}
        EXP['ana']['tau']='pg(002)'
        EXP['mono']={}
        EXP['mono']['tau']='pg(002)';
        EXP['ana']['mosaic']=30
        EXP['mono']['mosaic']=30
        EXP['sample']={}
        EXP['sample']['mosaic']=10
        EXP['sample']['vmosaic']=10
        EXP['hcol']=N.array([40, 10, 20, 80],'d')
        EXP['vcol']=N.array([120, 120, 120, 120],'d')
        EXP['infix']=-1 #positive for fixed incident energy
        EXP['efixed']=14.7
        EXP['method']=0
        setup=[EXP]  
        R0,RMS=mylattice.ResMatS(H,K,L,W,setup)
        print 'RMS'
        print RMS.transpose()[0]
#        mylattice.StandardSystem()
##    x1=N.array([1.0, 1.0], 'd'); y1=N.array([1.0, 1.0], 'd'); z1=N.array([1.0, 1.0], 'd'); x2=x1; y2=y1; z2=z1;
##    print 'scalar ', mylattice.scalar(x1,y1,z1,x2,y2,z2,'lattice')
##    print 'me ',mylattice.gtensor('lattice')
##    myinstrument=instrument();
##    print myinstrument.get_tau('pg(004)')
#    unittest.main()


import numpy as N
import pylab
from matplotlib.ticker import NullFormatter, MultipleLocator
from matplotlib.ticker import FormatStrFormatter
from matplotlib.ticker import MaxNLocator


if __name__=='__main__':
    if 1:
        #h k 0
        ax=pylab.subplot(3,4,1)
        #1 1 0
        x1=N.array([1])
        y1=N.array([1])

        pylab.plot(x1,y1,'bo',markersize=10,markerfacecolor='blue',markeredgecolor='blue')
        #-1 -1 0
        x1=N.array([-1])
        y1=N.array([-1])
        pylab.plot(x1,y1,'bo',markersize=10,markerfacecolor='blue',markeredgecolor='blue')
        #-1 2 0
        x1=N.array([-1])
        y1=N.array([2])
        pylab.plot(x1,y1,'bo',markersize=10,markerfacecolor='blue',markeredgecolor='blue')
        #1 -2 0
        x1=N.array([1])
        y1=N.array([-2])
        pylab.plot(x1,y1,'bo',markersize=10,markerfacecolor='blue',markeredgecolor='blue')
        #2 -1 0
        x1=N.array([2])
        y1=N.array([-1])
        pylab.plot(x1,y1,'bo',markersize=10,markerfacecolor='blue',markeredgecolor='blue')
        #-2 1 0
        x1=N.array([-2])
        y1=N.array([1])
        pylab.plot(x1,y1,'bo',markersize=10,markerfacecolor='blue',markeredgecolor='blue')
        pylab.xlabel('[1 0 0]')
        pylab.ylabel('[0 1 0]')
        #draw vertical line
        #x1=[0,0]
        #y1=[-2.5,2.5]
        #pylab.plot(x1,y1,linewidth=3.0)
        #draw horizontal line
        #x1=[-2.5,2.5]
        #y1=[0,0]
        #pylab.plot(x1,y1,linewidth=3.0,color='blue')

        #draw line 1 1 0
        x1=[1,-1]
        y1=[1,-1]
        pylab.plot(x1,y1,linewidth=3.0,color='blue')

        #draw line 1 2 0
        x1=[1,-1]
        y1=[-2,2]
        pylab.plot(x1,y1,linewidth=3.0,color='blue')

        #draw line 1 1 0
        x1=[-2,2]
        y1=[1,-1]
        pylab.plot(x1,y1,linewidth=3.0,color='blue')



        s=r'$\delta 2\bar{\delta} 0$'
        pylab.text(1.2,-2,s,fontsize=10)

        s=r'$2\delta \bar{\delta} 0$'
        pylab.text(1.8,-0.7,s,fontsize=10)

        s=r'$\delta \delta 0$'
        pylab.text(1,1.2,s,fontsize=10)

        s=r'$\bar{\delta} \bar{\delta} 0$'
        pylab.text(-1,-1.6,s,fontsize=10)

        s=r'$2\bar{\delta} \delta 0$'
        pylab.text(-2,1.2,s,fontsize=10)

        s=r'$\bar{\delta} 2\delta 0$'
        pylab.text(-.7,2.0,s,fontsize=10)

        pylab.axis([-3.5,3.5,-3.5,3.5])
        ax.yaxis.set_major_formatter(NullFormatter())
        ax.xaxis.set_major_formatter(NullFormatter())

    if 1:
        #h h l
        ax=pylab.subplot(3,4,2)

        # 1 1 0
        x1=N.array([4])
        y1=N.array([0])
        pylab.plot(x1,y1,'bo',markersize=10,markerfacecolor='blue',markeredgecolor='blue')
        s=r'$\delta \delta 0$'
        pylab.text(4.8,0.0,s,fontsize=10)

        # -1 -1 0
        x1=N.array([-4])
        y1=N.array([0])
        pylab.plot(x1,y1,'bo',markersize=10,markerfacecolor='blue',markeredgecolor='blue')
        s=r'$\bar{\delta} \bar{\delta} 0$'
        pylab.text(-6.8,0.0,s,fontsize=10)

        # -1 2 0
        x1=N.array([2])
        y1=N.array([0])
        pylab.plot(x1,y1,'bo',markersize=5,markerfacecolor='white')
        s=r'$\bar{\delta} 2\delta 0$'
        pylab.text(0.5,0.4,s,fontsize=10)

        # 2 -1 0
        #x1=N.array([2])
        #y1=N.array([0])
        #pylab.plot(x1,y1,'bo',markersize=10)
        s=r'$2\delta \bar{\delta} 0$'
        pylab.text(0.5,-1.4,s,fontsize=10)


        # -2 1 0
        x1=N.array([-2])
        y1=N.array([0])
        pylab.plot(x1,y1,'bo',markersize=5,markerfacecolor='white')
        s=r'$2\bar{\delta} \delta 0$'
        pylab.text(-3.5,0.4,s,fontsize=10)

        # 1 -2 0
        #x1=N.array([2])
        #y1=N.array([0])
        #pylab.plot(x1,y1,'bo',markersize=10)
        s=r'$\delta 2\bar{\delta} 0$'
        pylab.text(-3.5,-1.4,s,fontsize=10)



        pylab.axis([-8.0,8.0,-3.5,3.5])
        ax.yaxis.set_major_formatter(NullFormatter())
        ax.xaxis.set_major_formatter(NullFormatter())

        pylab.xlabel('[1 1 0]')
        pylab.ylabel('[0 0 1]')

        #ax=pylab.subplot(4,2,2)
    if 1:
        #h k h-k
        ax=pylab.subplot(3,2,2)
        # 1 1 0
        x1=N.array([4])
        y1=N.array([0])
        pylab.plot(x1,y1,'bo',markersize=20,markerfacecolor='blue',markeredgecolor='blue')
        s=r'$\delta \delta 0$'
        pylab.text(4.8,0.0,s,fontsize=20)

        # -1 -1 0
        x1=N.array([-4])
        y1=N.array([0])
        pylab.plot(x1,y1,'bo',markersize=20,markerfacecolor='blue',markeredgecolor='blue')
        s=r'$\bar{\delta} \bar{\delta} 0$'
        pylab.text(-6.8,0.0,s,fontsize=20)

        # -1 2 0
        x1=N.array([2])
        y1=N.array([4])
        pylab.plot(x1,y1,'bo',markersize=5,markerfacecolor='white')
        s=r'$\bar{\delta} 2\delta 0$'
        pylab.text(0.5,4.4,s,fontsize=20)


        # -2 1 0
        x1=N.array([-2])
        y1=N.array([4])
        pylab.plot(x1,y1,'bo',markersize=10,markerfacecolor='white')
        s=r'$2\bar{\delta} \delta 0$'
        pylab.text(-3.0,4.4,s,fontsize=20)


        # 2 -1 0
        x1=N.array([2])
        y1=N.array([-4])
        pylab.plot(x1,y1,'bo',markersize=10,markerfacecolor='gray')
        s=r'$2\delta \bar{\delta} 0$'
        pylab.text(0.5,-6.8,s,fontsize=20)


        # 1 -2 0
        x1=N.array([-2])
        y1=N.array([-4])
        pylab.plot(x1,y1,'bo',markersize=5,markerfacecolor='gray')
        s=r'$\delta 2\bar{\delta} 0$'
        pylab.text(-3.0,-6.8,s,fontsize=20)


        pylab.axis([-10.5,10.5,-10.5,10.5])
        ax.yaxis.set_major_formatter(NullFormatter())
        ax.xaxis.set_major_formatter(NullFormatter())

        pylab.xlabel('[1 1 0]')
        pylab.ylabel('[-1 1 -2]')

    if 1:
        pylab.subplot(3,2,3)
        pylab.subplot(3,2,4)
        pylab.subplot(3,2,5)
        pylab.subplot(3,2,6)
        pylab.subplots_adjust(wspace=0.5,hspace=0.6)


    if 1:
        pylab.show()






"""
Resolution Calculator for Triple Axis Spectrometer

This package provides GPU-accelerated resolution calculations using PyTorch.
It supports both NumPy (CPU) and PyTorch (GPU) backends for flexible deployment.

Key components:
- TASResolution: Resolution matrix calculator with GPU support
- ConvolutionCalculator: S(Q,w) convolution with resolution function
- Backends: NumPy and PyTorch computational backends

Example usage:
    from rescalculator_pytorch import TASResolution, get_backend
    from lattice_calculator import Lattice, Orientation

    # Initialize backend (auto-detects GPU)
    backend = get_backend('auto')

    # Set up lattice
    lattice = Lattice(a=5.0, b=5.0, c=5.0, ...)

    # Create resolution calculator
    res = TASResolution(lattice, backend='pytorch')

    # Calculate resolution matrices
    R0, RMS = res.ResMatS(H, K, L, W, EXP)
"""

__author__ = 'William Ratcliff'
__version__ = '2.0.0'

from .backends import get_backend, current_backend
from .rescalc_torch import TASResolution, ConvolutionCalculator

__all__ = [
    'TASResolution',
    'ConvolutionCalculator',
    'get_backend',
    'current_backend',
]

# Optional PyTorch-specific classes (require torch)
try:
    from .rescalc_batched import BatchedTASResolution
    __all__.append('BatchedTASResolution')
except ImportError:
    pass  # torch not installed



"""
Truly Batched PyTorch Resolution Calculator.

This module provides GPU-accelerated resolution calculations using BATCHED
tensor operations. Unlike the per-point loop version, this version constructs
all matrices for all Q-points simultaneously and uses batch matrix operations.

Key optimizations:
1. All matrices are 3D tensors of shape (n_pts, m, n)
2. Uses torch.bmm for batch matrix multiply
3. Uses torch.linalg.inv on batched matrices
4. Data stays on GPU until final result
"""

import numpy as np
import torch

try:
    from .backends import get_backend
except ImportError:
    from backends import get_backend

from lattice_calculator import modvec, scalar

# Physical constants
CONVERT1 = 0.4246609 * np.pi / 60 / 180  # arcmin to radians
CONVERT2 = 2.072  # meV to k^2 conversion


class BatchedTASResolution:
    """
    Batched Triple-Axis Spectrometer Resolution Calculator.

    This version uses fully vectorized/batched operations for GPU acceleration.
    All matrices are constructed as batched tensors and operations use
    torch.bmm (batch matrix multiply) for parallel computation.

    Parameters
    ----------
    lattice : Lattice
        Lattice calculator instance
    device : str, optional
        Device to use: 'auto', 'cuda', 'mps', or 'cpu'
    """

    def __init__(self, lattice, device='auto'):
        self.lattice = lattice

        # Set up device
        if device == 'auto':
            if torch.cuda.is_available():
                self.device = torch.device('cuda')
                self.dtype = torch.float64
            elif hasattr(torch.backends, 'mps') and torch.backends.mps.is_available():
                self.device = torch.device('mps')
                self.dtype = torch.float32  # MPS doesn't support float64
            else:
                self.device = torch.device('cpu')
                self.dtype = torch.float64
        else:
            self.device = torch.device(device)
            self.dtype = torch.float32 if 'mps' in str(device) else torch.float64

    def _to_tensor(self, x):
        """Convert to tensor on correct device."""
        if isinstance(x, torch.Tensor):
            return x.to(device=self.device, dtype=self.dtype)
        return torch.tensor(x, device=self.device, dtype=self.dtype)

    def _sign(self, x):
        """Vectorized sign function."""
        return torch.sign(x)

    def ResMat_batched(self, Q, W, EXP):
        """
        Compute resolution matrices for all Q,W points using batched operations.

        This is the main GPU-accelerated method. All matrix operations are batched
        across all points simultaneously.

        Parameters
        ----------
        Q : array_like
            Momentum transfer magnitudes, shape (npts,)
        W : array_like
            Energy transfer values, shape (npts,)
        EXP : list of dict
            Experiment configuration (uses first entry for common params)

        Returns
        -------
        R0 : tensor
            Resolution prefactors, shape (npts,)
        RM : tensor
            Resolution matrices, shape (npts, 4, 4)
        """
        npts = len(Q)

        # Convert inputs to tensors
        Q = self._to_tensor(Q)
        W = self._to_tensor(W)

        # Extract parameters from first EXP entry
        exp0 = EXP[0]
        mono = exp0['mono']
        ana = exp0['ana']
        sample = exp0['sample']

        # Collimations (radians)
        alpha = self._to_tensor(exp0['hcol']) * CONVERT1
        beta = self._to_tensor(exp0['vcol']) * CONVERT1

        # Mosaic spreads
        etam = float(mono['mosaic']) * CONVERT1
        etamv = float(mono.get('vmosaic', mono['mosaic'])) * CONVERT1
        etaa = float(ana['mosaic']) * CONVERT1
        etaav = float(ana.get('vmosaic', ana['mosaic'])) * CONVERT1

        # Fixed energy and direction
        infin = exp0.get('infin', -1)
        efixed = float(exp0['efixed'])
        epm = exp0.get('dir1', 1)
        ep = exp0.get('dir2', 1)

        # Tau values
        tau_list = {
            'pg(002)': 1.87325, 'pg(004)': 3.74650,
            'ge(111)': 1.92366, 'ge(220)': 3.14131,
            'ge(311)': 3.68351, 'be(002)': 3.50702,
            'pg(110)': 5.49806
        }
        taum = tau_list[mono['tau']]
        taua = tau_list[ana['tau']]

        # Arms
        arms = exp0.get('arms', [1.0, 1.0, 1.0, 1.0, 1.0])
        L0, L1, L2, L3 = float(arms[0]), float(arms[2]), float(arms[3]), float(arms[4])

        # Focusing
        monorv = float(mono.get('rv', 1e6))
        monorh = float(mono.get('rh', 1e6))
        anarv = float(ana.get('rv', 1e6))
        anarh = float(ana.get('rh', 1e6))
        horifoc = exp0.get('horifoc', -1)
        moncor = exp0.get('moncor', 1)

        # ========== VECTORIZED ENERGY/ANGLE CALCULATIONS ==========
        # Energies for all points
        if infin > 0:
            ei = torch.full((npts,), efixed, device=self.device, dtype=self.dtype)
            ef = efixed - W
        else:
            ei = efixed + W
            ef = torch.full((npts,), efixed, device=self.device, dtype=self.dtype)

        # Wave vectors (npts,)
        ki = torch.sqrt(ei / CONVERT2)
        kf = torch.sqrt(ef / CONVERT2)

        # Bragg angles (npts,)
        sign_epm = 1 if epm > 0 else -1
        sign_ep = 1 if ep > 0 else -1
        thetam = torch.arcsin(taum / (2 * ki)) * sign_epm
        thetaa = torch.arcsin(taua / (2 * kf)) * sign_ep

        # Sample angles (npts,)
        cos_s2theta = (ki**2 + kf**2 - Q**2) / (2 * ki * kf)
        cos_s2theta = torch.clamp(cos_s2theta, -1.0, 1.0)
        s2theta = -torch.arccos(cos_s2theta)
        thetas = s2theta / 2
        phi = torch.arctan2(-kf * torch.sin(s2theta), ki - kf * torch.cos(s2theta))

        # ========== BUILD BATCHED MATRICES ==========
        # G matrix - diagonal, same for all points: shape (8,)
        G_diag = 1.0 / torch.stack([
            alpha[0], alpha[1], beta[0], beta[1],
            alpha[2], alpha[3], beta[2], beta[3]
        ])**2
        # Expand to (npts, 8, 8) diagonal
        G = torch.diag(G_diag).unsqueeze(0).expand(npts, -1, -1).clone()

        # F matrix - diagonal, same for all points
        F_diag = self._to_tensor([1.0/etam**2, 1.0/etamv**2, 1.0/etaa**2, 1.0/etaav**2])
        F = torch.diag(F_diag).unsqueeze(0).expand(npts, -1, -1).clone()

        # A matrix - varies with ki, kf, thetam, thetaa: shape (npts, 6, 8)
        A = torch.zeros((npts, 6, 8), device=self.device, dtype=self.dtype)
        A[:, 0, 0] = ki / (2 * torch.tan(thetam))
        A[:, 0, 1] = -ki / (2 * torch.tan(thetam))
        A[:, 3, 4] = kf / (2 * torch.tan(thetaa))
        A[:, 3, 5] = -kf / (2 * torch.tan(thetaa))
        A[:, 1, 1] = ki
        A[:, 2, 3] = ki
        A[:, 4, 4] = kf
        A[:, 5, 6] = kf

        # C matrix - varies with thetam, thetaa: shape (npts, 4, 8)
        C = torch.zeros((npts, 4, 8), device=self.device, dtype=self.dtype)
        C[:, 0, 0] = 0.5
        C[:, 0, 1] = 0.5
        C[:, 2, 4] = 0.5
        C[:, 2, 5] = 0.5
        C[:, 1, 2] = 1.0 / (2 * torch.sin(thetam))
        C[:, 1, 3] = -1.0 / (2 * torch.sin(thetam))
        C[:, 3, 6] = 1.0 / (2 * torch.sin(thetaa))
        C[:, 3, 7] = -1.0 / (2 * torch.sin(thetaa))

        # B matrix - varies with phi, s2theta, ki, kf: shape (npts, 4, 6)
        B = torch.zeros((npts, 4, 6), device=self.device, dtype=self.dtype)
        B[:, 0, 0] = torch.cos(phi)
        B[:, 0, 1] = torch.sin(phi)
        B[:, 0, 3] = -torch.cos(phi - s2theta)
        B[:, 0, 4] = -torch.sin(phi - s2theta)
        B[:, 1, 0] = -torch.sin(phi)
        B[:, 1, 1] = torch.cos(phi)
        B[:, 1, 3] = torch.sin(phi - s2theta)
        B[:, 1, 4] = -torch.cos(phi - s2theta)
        B[:, 2, 2] = 1.0
        B[:, 2, 5] = -1.0
        B[:, 3, 0] = 2 * CONVERT2 * ki
        B[:, 3, 3] = -2 * CONVERT2 * kf

        # ========== BATCHED MATRIX OPERATIONS ==========
        # Cooper-Nathans calculation: M = B @ A @ inv(G + C.T @ F @ C) @ A.T @ B.T

        # C.T @ F @ C: (npts, 8, 4) @ (npts, 4, 4) @ (npts, 4, 8) -> (npts, 8, 8)
        CT = C.transpose(-2, -1)  # (npts, 8, 4)
        CtFC = torch.bmm(torch.bmm(CT, F), C)  # (npts, 8, 8)

        # G + C.T @ F @ C
        GpCtFC = G + CtFC  # (npts, 8, 8)

        # Inverse: (npts, 8, 8) -> (npts, 8, 8)
        GpCtFC_inv = torch.linalg.inv(GpCtFC)

        # A @ inv(...) @ A.T: (npts, 6, 8) @ (npts, 8, 8) @ (npts, 8, 6) -> (npts, 6, 6)
        AT = A.transpose(-2, -1)  # (npts, 8, 6)
        HF = torch.bmm(torch.bmm(A, GpCtFC_inv), AT)  # (npts, 6, 6)

        # Horizontal focusing correction if needed
        if horifoc > 0:
            HF_inv = torch.linalg.inv(HF)
            HF_inv[:, 4, 4] = (1.0 / (kf * alpha[2]))**2
            HF_inv[:, 4, 3] = 0
            HF_inv[:, 3, 4] = 0
            HF_inv[:, 3, 3] = (torch.tan(thetaa) / (etaa * kf))**2
            HF = torch.linalg.inv(HF_inv)

        # B @ HF @ B.T: (npts, 4, 6) @ (npts, 6, 6) @ (npts, 6, 4) -> (npts, 4, 4)
        BT = B.transpose(-2, -1)  # (npts, 6, 4)
        Minv = torch.bmm(torch.bmm(B, HF), BT)  # (npts, 4, 4)

        # Final inversion
        M = torch.linalg.inv(Minv)  # (npts, 4, 4)

        # Reorder to RM format (swap indices 2 and 3)
        # M indices: 0=Qx, 1=Qy, 2=Qz, 3=E
        # RM indices: 0=Qx, 1=Qy, 2=E, 3=Qz
        RM = torch.zeros_like(M)
        idx = [0, 1, 3, 2]  # Reordering
        for i in range(4):
            for j in range(4):
                RM[:, i, j] = M[:, idx[i], idx[j]]

        # ========== PREFACTOR CALCULATION ==========
        Rm = ki**3 / torch.tan(thetam)  # (npts,)
        Ra = kf**3 / torch.tan(thetaa)  # (npts,)

        pi = np.pi
        det_F = torch.linalg.det(F)  # (npts,) but all same value
        det_GpCtFC = torch.linalg.det(GpCtFC)  # (npts,)

        R0 = Rm * Ra * (2*pi)**4 / (64 * pi**2 * torch.sin(thetam) * torch.sin(thetaa)) \
             * torch.sqrt(det_F / det_GpCtFC)

        # Monitor correction
        if moncor == 1:
            g = G[:, 0:4, 0:4]  # (npts, 4, 4)
            f = F[:, 0:2, 0:2]  # (npts, 2, 2)
            c = C[:, 0:2, 0:4]  # (npts, 2, 4)
            ct = c.transpose(-2, -1)  # (npts, 4, 2)
            ctfc = torch.bmm(torch.bmm(ct, f), c)  # (npts, 4, 4)
            det_f = torch.linalg.det(f)
            det_gctfc = torch.linalg.det(g + ctfc)

            Rmon = Rm * (2*pi)**2 / (8*pi * torch.sin(thetam)) * torch.sqrt(det_f / det_gctfc)
            R0 = R0 / Rmon * ki

        # Chesser-Axe normalization
        det_RM = torch.linalg.det(RM)
        R0 = R0 / (2*pi)**2 * torch.sqrt(det_RM)

        # kf/ki factor
        R0 = R0 * kf / ki

        # Sample mosaic correction
        if 'mosaic' in sample:
            etas = float(sample['mosaic']) * CONVERT1
            etasv = float(sample.get('vmosaic', sample['mosaic'])) * CONVERT1

            R0 = R0 / torch.sqrt((1 + (Q * etas)**2 * RM[:, 3, 3]) *
                                 (1 + (Q * etasv)**2 * RM[:, 1, 1]))

            RM_inv = torch.linalg.inv(RM)
            RM_inv[:, 1, 1] = RM_inv[:, 1, 1] + Q**2 * etas**2
            RM_inv[:, 3, 3] = RM_inv[:, 3, 3] + Q**2 * etasv**2
            RM = torch.linalg.inv(RM_inv)

        return R0, RM

    def ResMatS(self, H, K, L, W, EXP):
        """
        Compute resolution matrices in sample coordinate system.

        Parameters
        ----------
        H, K, L : array_like
            Miller indices
        W : array_like
            Energy transfers
        EXP : list of dict
            Experiment configuration

        Returns
        -------
        R0 : ndarray
            Prefactors
        RMS : ndarray
            Resolution matrices in sample coords, shape (4, 4, npts)
        """
        H = np.atleast_1d(H)
        K = np.atleast_1d(K)
        L = np.atleast_1d(L)
        W = np.atleast_1d(W)

        npts = len(H)
        self.lattice.npts = npts

        x = self.lattice.x
        y = self.lattice.y

        Q = modvec(H, K, L, 'latticestar', self.lattice)

        # Unit vector along Q
        uq = np.zeros((3, npts))
        uq[0, :] = H / Q
        uq[1, :] = K / Q
        uq[2, :] = L / Q

        # Scalar products
        xq = scalar(x[0, :], x[1, :], x[2, :], uq[0, :], uq[1, :], uq[2, :], 'latticestar', self.lattice)
        yq = scalar(y[0, :], y[1, :], y[2, :], uq[0, :], uq[1, :], uq[2, :], 'latticestar', self.lattice)

        # Build transformation matrix (npts, 4, 4)
        tmat = self._to_tensor(np.zeros((npts, 4, 4)))
        tmat[:, 3, 3] = 1
        tmat[:, 2, 2] = 1
        tmat[:, 0, 0] = self._to_tensor(xq)
        tmat[:, 0, 1] = self._to_tensor(yq)
        tmat[:, 1, 1] = self._to_tensor(xq)
        tmat[:, 1, 0] = -self._to_tensor(yq)

        # Get resolution matrices
        R0, RM = self.ResMat_batched(Q, W, EXP)

        # Transform: RMS = tmat.T @ RM @ tmat (batched)
        tmatT = tmat.transpose(-2, -1)
        RMS = torch.bmm(torch.bmm(tmatT, RM), tmat)

        # Convert to numpy and transpose for legacy format (4, 4, npts)
        R0_np = R0.cpu().numpy()
        RMS_np = RMS.cpu().numpy().transpose(1, 2, 0)

        return R0_np, RMS_np



"""
PyTorch-accelerated Triple-Axis Spectrometer Resolution Calculator.

This module provides GPU-accelerated resolution calculations using PyTorch.
The calculation follows the Cooper-Nathans formalism with optional Popovici method.
"""

import numpy as np

try:
    from .backends import get_backend, current_backend
except ImportError:
    from backends import get_backend, current_backend

# Import lattice_calculator functions
from lattice_calculator import modvec, scalar

# Physical constants
CONVERT1 = 0.4246609 * np.pi / 60 / 180  # arcmin to radians factor
CONVERT2 = 2.072  # meV to k^2 conversion


class TASResolution:
    """
    Triple-Axis Spectrometer Resolution Calculator with PyTorch acceleration.

    This class computes resolution matrices and prefactors for TAS experiments
    using either CPU (NumPy) or GPU (PyTorch) backends.

    Parameters
    ----------
    lattice : LatticeCalculator
        Lattice calculator instance with crystal and orientation info
    backend : str, optional
        Backend to use: 'auto', 'numpy', or 'pytorch'
    """

    def __init__(self, lattice, backend='auto'):
        self.lattice = lattice
        self._backend = get_backend(backend)
        self.B = self._backend  # shorthand

    @property
    def backend(self):
        return self._backend

    def set_backend(self, backend_type):
        """Change the computational backend."""
        self._backend = get_backend(backend_type)
        self.B = self._backend

    def _sign(self, x):
        """Sign function that handles scalars properly."""
        if hasattr(x, '__len__'):
            return self.B.sign(x)
        return 1 if x > 0 else (-1 if x < 0 else 0)

    def ResMat_vectorized(self, Q, W, EXP):
        """
        Compute resolution matrices for multiple Q,W points - vectorized version.

        This is the main GPU-accelerated method that computes resolution matrices
        for all points simultaneously using batch matrix operations.

        Parameters
        ----------
        Q : array_like
            Momentum transfer magnitudes, shape (npts,)
        W : array_like
            Energy transfer values, shape (npts,)
        EXP : list of dict
            Experiment configuration dictionaries

        Returns
        -------
        R0 : array
            Resolution prefactors, shape (npts,)
        RM : array
            Resolution matrices, shape (4, 4, npts)
        """
        B = self.B
        npts = len(EXP)

        # Convert inputs to backend arrays
        Q = B.to_array(Q)
        W = B.to_array(W)

        # Pre-allocate output arrays
        RM = B.zeros((4, 4, npts))
        R0 = B.zeros((npts,))

        # Extract parameters from EXP (vectorize where possible)
        # For now, we handle the common case where all EXP entries are identical
        # except for position-dependent quantities

        # Get common parameters from first EXP entry
        exp0 = EXP[0]
        mono = exp0['mono']
        ana = exp0['ana']
        sample = exp0['sample']

        # Collimations (convert to radians) - keep as numpy for matrix calculations
        alpha = np.asarray(exp0['hcol'], dtype=np.float64) * CONVERT1
        beta = np.asarray(exp0['vcol'], dtype=np.float64) * CONVERT1

        # Mosaic spreads - keep as python floats
        etam = float(mono['mosaic']) * CONVERT1
        etamv = float(mono.get('vmosaic', mono['mosaic'])) * CONVERT1
        etaa = float(ana['mosaic']) * CONVERT1
        etaav = float(ana.get('vmosaic', ana['mosaic'])) * CONVERT1

        # Fixed energy and direction
        infin = exp0.get('infin', -1)
        efixed = exp0['efixed']
        epm = exp0.get('dir1', 1)
        ep = exp0.get('dir2', 1)

        # Get tau values for monochromator and analyzer
        tau_list = {
            'pg(002)': 1.87325, 'pg(004)': 3.74650,
            'ge(111)': 1.92366, 'ge(220)': 3.14131,
            'ge(311)': 3.68351, 'be(002)': 3.50702,
            'pg(110)': 5.49806
        }
        taum = tau_list[mono['tau']]
        taua = tau_list[ana['tau']]

        # Arms lengths
        arms = exp0.get('arms', [1.0, 1.0, 1.0, 1.0, 1.0])
        L0 = arms[0]
        L1 = arms[2] if len(arms) > 2 else 1.0
        L2 = arms[3] if len(arms) > 3 else 1.0
        L3 = arms[4] if len(arms) > 4 else 1.0
        L1mon = arms[4] if len(arms) > 4 else L1

        # Focusing parameters
        monorv = mono.get('rv', 1e6)
        monorh = mono.get('rh', 1e6)
        anarv = ana.get('rv', 1e6)
        anarh = ana.get('rh', 1e6)

        horifoc = exp0.get('horifoc', -1)
        method = exp0.get('method', 0)
        moncor = exp0.get('moncor', 1)

        # Calculate energies for all points
        ei = B.where(infin > 0, efixed * B.ones(npts), efixed + W)
        ef = B.where(infin > 0, efixed - W, efixed * B.ones(npts))

        # Wave vectors
        ki = B.sqrt(ei / CONVERT2)
        kf = B.sqrt(ef / CONVERT2)

        # Bragg angles
        thetam = B.arcsin(taum / (2 * ki)) * self._sign(epm)
        thetaa = B.arcsin(taua / (2 * kf)) * self._sign(ep)

        # Sample angle
        s2theta = -B.arccos((ki**2 + kf**2 - Q**2) / (2 * ki * kf))
        thetas = s2theta / 2
        phi = B.arctan2(-kf * B.sin(s2theta), ki - kf * B.cos(s2theta))

        # Handle guide divergences (negative alpha/beta values)
        pi = np.pi
        for i in range(4):
            if alpha[i] < 0:
                alpha[i] = -alpha[i] * 2 * 0.427 / B.to_numpy(ki[0]) * pi / 180
            if beta[i] < 0:
                beta[i] = -beta[i] * 2 * 0.427 / B.to_numpy(ki[0]) * pi / 180

        # Build matrices for all points simultaneously using batch operations
        # We'll compute the resolution matrix for each point

        for ind in range(npts):
            # Extract scalar values for this point
            ki_i = ki[ind] if hasattr(ki, '__getitem__') else ki
            kf_i = kf[ind] if hasattr(kf, '__getitem__') else kf
            thetam_i = thetam[ind] if hasattr(thetam, '__getitem__') else thetam
            thetaa_i = thetaa[ind] if hasattr(thetaa, '__getitem__') else thetaa
            thetas_i = thetas[ind] if hasattr(thetas, '__getitem__') else thetas
            s2theta_i = s2theta[ind] if hasattr(s2theta, '__getitem__') else s2theta
            phi_i = phi[ind] if hasattr(phi, '__getitem__') else phi
            q_i = Q[ind] if hasattr(Q, '__getitem__') else Q
            w_i = W[ind] if hasattr(W, '__getitem__') else W

            # Convert to numpy for intermediate calculations if using PyTorch
            ki_n = float(B.to_numpy(ki_i)) if hasattr(ki_i, 'numpy') else float(ki_i)
            kf_n = float(B.to_numpy(kf_i)) if hasattr(kf_i, 'numpy') else float(kf_i)
            thetam_n = float(B.to_numpy(thetam_i)) if hasattr(thetam_i, 'numpy') else float(thetam_i)
            thetaa_n = float(B.to_numpy(thetaa_i)) if hasattr(thetaa_i, 'numpy') else float(thetaa_i)
            thetas_n = float(B.to_numpy(thetas_i)) if hasattr(thetas_i, 'numpy') else float(thetas_i)
            s2theta_n = float(B.to_numpy(s2theta_i)) if hasattr(s2theta_i, 'numpy') else float(s2theta_i)
            phi_n = float(B.to_numpy(phi_i)) if hasattr(phi_i, 'numpy') else float(phi_i)
            q_n = float(B.to_numpy(q_i)) if hasattr(q_i, 'numpy') else float(q_i)

            # Build G matrix (collimation)
            G_diag = 1.0 / np.array([
                alpha[0], alpha[1], beta[0], beta[1],
                alpha[2], alpha[3], beta[2], beta[3]
            ])**2
            G = np.diag(G_diag)

            # Build F matrix (mosaic)
            F_diag = 1.0 / np.array([etam, etamv, etaa, etaav])**2
            F = np.diag(F_diag)

            # Build A matrix
            A = np.zeros((6, 8))
            A[0, 0] = ki_n / 2 / np.tan(thetam_n)
            A[0, 1] = -A[0, 0]
            A[3, 4] = kf_n / 2 / np.tan(thetaa_n)
            A[3, 5] = -A[3, 4]
            A[1, 1] = ki_n
            A[2, 3] = ki_n
            A[4, 4] = kf_n
            A[5, 6] = kf_n

            # Build C matrix
            C = np.zeros((4, 8))
            C[0, 0] = 0.5
            C[0, 1] = 0.5
            C[2, 4] = 0.5
            C[2, 5] = 0.5
            C[1, 2] = 1.0 / (2 * np.sin(thetam_n))
            C[1, 3] = -C[1, 2]
            C[3, 6] = 1.0 / (2 * np.sin(thetaa_n))
            C[3, 7] = -C[3, 6]

            # Build B matrix
            B_mat = np.zeros((4, 6))
            B_mat[0, 0] = np.cos(phi_n)
            B_mat[0, 1] = np.sin(phi_n)
            B_mat[0, 3] = -np.cos(phi_n - s2theta_n)
            B_mat[0, 4] = -np.sin(phi_n - s2theta_n)
            B_mat[1, 0] = -B_mat[0, 1]
            B_mat[1, 1] = B_mat[0, 0]
            B_mat[1, 3] = -B_mat[0, 4]
            B_mat[1, 4] = B_mat[0, 3]
            B_mat[2, 2] = 1.0
            B_mat[2, 5] = -1.0
            B_mat[3, 0] = 2 * CONVERT2 * ki_n
            B_mat[3, 3] = -2 * CONVERT2 * kf_n

            # Cooper-Nathans calculation
            HF_int = np.linalg.inv(G + C.T @ F @ C)
            HF = A @ HF_int @ A.T

            # Horizontal focusing
            if horifoc > 0:
                HF = np.linalg.inv(HF)
                HF[4, 4] = (1.0 / (kf_n * alpha[2]))**2
                HF[4, 3] = 0
                HF[3, 4] = 0
                HF[3, 3] = (np.tan(thetaa_n) / (etaa * kf_n))**2
                HF = np.linalg.inv(HF)

            Minv = B_mat @ HF @ B_mat.T
            M = np.linalg.inv(Minv)

            # Reorder matrix elements
            RM_ = np.zeros((4, 4))
            RM_[0, 0] = M[0, 0]
            RM_[1, 0] = M[1, 0]
            RM_[0, 1] = M[0, 1]
            RM_[1, 1] = M[1, 1]
            RM_[0, 2] = M[0, 3]
            RM_[2, 0] = M[3, 0]
            RM_[2, 2] = M[3, 3]
            RM_[2, 1] = M[3, 1]
            RM_[1, 2] = M[1, 3]
            RM_[0, 3] = M[0, 2]
            RM_[3, 0] = M[2, 0]
            RM_[3, 3] = M[2, 2]
            RM_[3, 1] = M[2, 1]
            RM_[1, 3] = M[1, 2]
            RM_[3, 2] = M[2, 3]
            RM_[2, 3] = M[3, 2]

            # Calculate prefactor
            Rm = ki_n**3 / np.tan(thetam_n)
            Ra = kf_n**3 / np.tan(thetaa_n)

            R0_ = Rm * Ra * (2*pi)**4 / (64 * pi**2 * np.sin(thetam_n) * np.sin(thetaa_n)) \
                  * np.sqrt(np.linalg.det(F) / np.linalg.det(G + C.T @ F @ C))

            # Monitor correction
            if moncor == 1:
                g = G[0:4, 0:4]
                f = F[0:2, 0:2]
                c = C[0:2, 0:4]
                Rmon = Rm * (2*pi)**2 / (8*pi * np.sin(thetam_n)) \
                       * np.sqrt(np.linalg.det(f) / np.linalg.det(g + c.T @ f @ c))
                R0_ = R0_ / Rmon
                R0_ = R0_ * ki_n

            # Chesser-Axe normalization
            R0_ = R0_ / (2*pi)**2 * np.sqrt(np.linalg.det(RM_))

            # kf/ki factor
            R0_ = R0_ * kf_n / ki_n

            # Sample mosaic correction
            if 'mosaic' in sample:
                etas = sample['mosaic'] * CONVERT1
                etasv = sample.get('vmosaic', sample['mosaic']) * CONVERT1
                R0_ = R0_ / np.sqrt((1 + (q_n * etas)**2 * RM_[3, 3]) *
                                   (1 + (q_n * etasv)**2 * RM_[1, 1]))
                Minv = np.linalg.inv(RM_)
                Minv[1, 1] = Minv[1, 1] + q_n**2 * etas**2
                Minv[3, 3] = Minv[3, 3] + q_n**2 * etasv**2
                RM_ = np.linalg.inv(Minv)

            # Store results
            R0[ind] = R0_
            RM[:, :, ind] = B.to_array(RM_)

        return R0, RM

    def ResMatS(self, H, K, L, W, EXP):
        """
        Compute resolution matrices in sample coordinate system.

        Parameters
        ----------
        H, K, L : array_like
            Miller indices of Q points
        W : array_like
            Energy transfer values
        EXP : list of dict
            Experiment configuration

        Returns
        -------
        R0 : array
            Resolution prefactors
        RMS : array
            Resolution matrices in sample coordinates
        """
        B = self.B

        x = self.lattice.x
        y = self.lattice.y
        z = self.lattice.z

        Q = modvec(H, K, L, 'latticestar', self.lattice)
        npts = self.lattice.npts

        # Unit vector along Q
        uq = B.zeros((3, npts))
        uq[0, :] = B.to_array(H) / B.to_array(Q)
        uq[1, :] = B.to_array(K) / B.to_array(Q)
        uq[2, :] = B.to_array(L) / B.to_array(Q)

        # Scalar products
        xq = scalar(x[0, :], x[1, :], x[2, :],
                    B.to_numpy(uq[0, :]), B.to_numpy(uq[1, :]), B.to_numpy(uq[2, :]),
                    'latticestar', self.lattice)
        yq = scalar(y[0, :], y[1, :], y[2, :],
                    B.to_numpy(uq[0, :]), B.to_numpy(uq[1, :]), B.to_numpy(uq[2, :]),
                    'latticestar', self.lattice)

        # Transformation matrix
        tmat = B.zeros((4, 4, npts))
        tmat[3, 3, :] = 1
        tmat[2, 2, :] = 1
        tmat[0, 0, :] = B.to_array(xq)
        tmat[0, 1, :] = B.to_array(yq)
        tmat[1, 1, :] = B.to_array(xq)
        tmat[1, 0, :] = -B.to_array(yq)

        # Get resolution matrices
        R0, RM = self.ResMat_vectorized(Q, W, EXP)

        # Transform to sample coordinates
        RMS = B.zeros((4, 4, npts))
        for i in range(npts):
            tmat_i = B.to_numpy(tmat[:, :, i])
            RM_i = B.to_numpy(RM[:, :, i])
            RMS[:, :, i] = B.to_array(tmat_i.T @ RM_i @ tmat_i)

        return B.to_numpy(R0).T, RMS


class ConvolutionCalculator:
    """
    GPU-accelerated convolution of S(Q,w) with resolution function.

    This class handles the numerical convolution of a user-defined cross-section
    with the TAS resolution function using vectorized operations.

    Parameters
    ----------
    resolution : TASResolution
        Resolution calculator instance
    backend : str, optional
        Backend to use
    """

    def __init__(self, resolution, backend='auto'):
        self.resolution = resolution
        self._backend = get_backend(backend)
        self.B = self._backend

    def convolve(self, sqw_func, pref_func, H, K, L, W, EXP, p,
                 method='fixed', accuracy=(7, 0)):
        """
        Convolve S(Q,w) with resolution function.

        This is the main GPU-accelerated convolution routine. It samples the
        cross-section at multiple points within the resolution ellipsoid and
        weights by the resolution function.

        Parameters
        ----------
        sqw_func : callable
            S(Q,w) function: sqw(H, K, L, W, p) -> array of shape (modes, npts)
        pref_func : callable or None
            Prefactor function for polarization, etc.
        H, K, L : array_like
            Miller indices of scan points
        W : array_like
            Energy transfers
        EXP : list of dict
            Experiment configuration
        p : dict
            Parameters for sqw_func
        method : str
            'fixed' for Gaussian quadrature, 'mc' for Monte Carlo
        accuracy : tuple
            (M0, M1) accuracy parameters for fixed method

        Returns
        -------
        conv : array
            Convolved intensities
        """
        B = self.B
        pi = np.pi

        H = B.to_array(H)
        K = B.to_array(K)
        L = B.to_array(L)
        W = B.to_array(W)

        npts = len(B.to_numpy(H))

        # Get coordinate system vectors
        xvec = self.resolution.lattice.x
        yvec = self.resolution.lattice.y
        zvec = self.resolution.lattice.z

        # Calculate resolution matrices
        R0, RMS = self.resolution.ResMatS(B.to_numpy(H), B.to_numpy(K), B.to_numpy(L),
                                          B.to_numpy(W), EXP)

        # Extract matrix elements
        Mzz = B.to_array(RMS[3, 3, :])
        Mww = B.to_array(RMS[2, 2, :])
        Mxx = B.to_array(RMS[0, 0, :])
        Mxy = B.to_array(RMS[0, 1, :])
        Mxw = B.to_array(RMS[0, 2, :])
        Myy = B.to_array(RMS[1, 1, :])
        Myw = B.to_array(RMS[1, 2, :])

        # Compute effective variances
        Mxx = Mxx - Mxw**2 / Mww
        Mxy = Mxy - Mxw * Myw / Mww
        Myy = Myy - Myw**2 / Mww
        MMxx = Mxx - Mxy**2 / Myy
        detM = MMxx * Myy * Mzz * Mww

        # Transformation coefficients
        tqz = 1.0 / B.sqrt(Mzz)
        tqx = 1.0 / B.sqrt(MMxx)
        tqyy = 1.0 / B.sqrt(Myy)
        tqyx = -Mxy / Myy / B.sqrt(MMxx)
        tqww = 1.0 / B.sqrt(Mww)
        tqwy = -Myw / Mww / B.sqrt(Myy)
        tqwx = -(Mxw / Mww - Myw / Mww * Mxy / Myy) / B.sqrt(MMxx)

        # Test sqw function to get number of modes
        H_np = B.to_numpy(H)
        K_np = B.to_numpy(K)
        L_np = B.to_numpy(L)
        W_np = B.to_numpy(W)

        test_sqw = sqw_func(H_np, K_np, L_np, W_np, p)
        modes = test_sqw.shape[0]

        # Get prefactors
        if pref_func is None:
            prefactor = np.ones((modes, npts), dtype=np.float64)
            bgr = 0
        else:
            prefactor, bgr = pref_func(H_np, K_np, L_np, W_np, self.resolution, p)

        if method == 'fixed':
            conv = self._convolve_fixed(
                sqw_func, H, K, L, W, p, modes, npts,
                tqx, tqyy, tqyx, tqz, tqwx, tqwy, tqww,
                xvec, yvec, zvec, detM, prefactor, accuracy
            )
        else:
            raise NotImplementedError(f"Method '{method}' not yet implemented")

        # Apply R0 and background
        R0_tensor = B.to_array(R0)
        conv = conv * R0_tensor + bgr

        return B.to_numpy(conv)

    def _convolve_fixed(self, sqw_func, H, K, L, W, p, modes, npts,
                        tqx, tqyy, tqyx, tqz, tqwx, tqwy, tqww,
                        xvec, yvec, zvec, detM, prefactor, accuracy):
        """
        Fixed-point Gaussian quadrature convolution.

        This version is optimized for GPU by:
        1. Pre-computing all sampling points on GPU
        2. Vectorizing the cross-section evaluation
        3. Using batch matrix operations for transformations
        """
        B = self.B
        pi = np.pi

        M = accuracy
        step1 = pi / (2 * M[0] + 1)
        step2 = pi / (2 * M[1] + 1) if M[1] > 0 else pi

        # Create sampling grids
        n1 = 2 * M[0] + 1
        n2 = max(2 * M[1] + 1, 1)

        dd1 = B.linspace(-pi/2 + step1/2, pi/2 - step1/2, n1)
        dd2 = B.linspace(-pi/2 + step2/2, pi/2 - step2/2, n2) if M[1] > 0 else B.to_array([0.0])

        # Create 3D meshgrid for x, y, w sampling
        cx, cy, cw = B.meshgrid(dd1, dd1, dd1, indexing='ij')

        # Flatten and compute tan values
        cx_flat = B.reshape(cx, (-1,))
        cy_flat = B.reshape(cy, (-1,))
        cw_flat = B.reshape(cw, (-1,))

        tx = B.tan(cx_flat)
        ty = B.tan(cy_flat)
        tw = B.tan(cw_flat)
        tz = B.tan(dd2)

        # Compute weights
        norm = (1 + tx**2) * (1 + ty**2) * (1 + tw**2) * B.exp(-0.5 * (tx**2 + ty**2 + tw**2))
        normz = B.exp(-0.5 * tz**2) * (1 + tz**2)

        n_samples = len(B.to_numpy(tx))
        n_z = len(B.to_numpy(tz))

        # Initialize convolution array
        convs = B.zeros((modes, npts))
        conv = B.zeros((npts,))

        # Convert to numpy for sqw evaluation
        H_np = B.to_numpy(H)
        K_np = B.to_numpy(K)
        L_np = B.to_numpy(L)
        W_np = B.to_numpy(W)
        tx_np = B.to_numpy(tx)
        ty_np = B.to_numpy(ty)
        tw_np = B.to_numpy(tw)
        tz_np = B.to_numpy(tz)
        norm_np = B.to_numpy(norm)
        normz_np = B.to_numpy(normz)
        tqx_np = B.to_numpy(tqx)
        tqyy_np = B.to_numpy(tqyy)
        tqyx_np = B.to_numpy(tqyx)
        tqz_np = B.to_numpy(tqz)
        tqwx_np = B.to_numpy(tqwx)
        tqwy_np = B.to_numpy(tqwy)
        tqww_np = B.to_numpy(tqww)

        convs_np = np.zeros((modes, npts))

        # Main convolution loop
        for iz in range(n_z):
            for i in range(npts):
                # Compute displacements in Q-E space
                dQ1 = tqx_np[i] * tx_np
                dQ2 = tqyy_np[i] * ty_np + tqyx_np[i] * tx_np
                dW = tqwx_np[i] * tx_np + tqwy_np[i] * ty_np + tqww_np[i] * tw_np
                dQ4 = tqz_np[i] * tz_np[iz]

                # Compute displaced Q points
                H1 = H_np[i] + dQ1 * xvec[0, i] + dQ2 * yvec[0, i] + dQ4 * zvec[0, i]
                K1 = K_np[i] + dQ1 * xvec[1, i] + dQ2 * yvec[1, i] + dQ4 * zvec[1, i]
                L1 = L_np[i] + dQ1 * xvec[2, i] + dQ2 * yvec[2, i] + dQ4 * zvec[2, i]
                W1 = W_np[i] + dW

                # Evaluate cross-section
                myint = sqw_func(H1, K1, L1, W1, p)

                # Accumulate weighted contribution
                for j in range(modes):
                    if myint.ndim > 1:
                        add = myint[j, :] * norm_np * normz_np[iz]
                    else:
                        add = myint[j] * norm_np * normz_np[iz]
                    convs_np[j, i] += np.sum(add)

        # Apply prefactor and sum modes
        conv_np = np.sum(convs_np * prefactor, axis=0)

        # Apply normalization factor
        detM_np = B.to_numpy(detM)
        factor = step1**3 * step2 / np.sqrt(detM_np)
        conv_np = conv_np * factor

        # Correction factors for low accuracy
        if M[1] == 0:
            conv_np = conv_np * 0.79788
        if M[0] == 0:
            conv_np = conv_np * 0.79788**3

        return B.to_array(conv_np)

    def convolve_batched(self, sqw_func, H, K, L, W, EXP, p,
                         method='fixed', accuracy=(7, 0), batch_size=1000):
        """
        Batched convolution for very large grids.

        Splits the calculation into batches to manage GPU memory.

        Parameters
        ----------
        batch_size : int
            Number of Q points per batch
        """
        B = self.B

        H = np.asarray(H)
        K = np.asarray(K)
        L = np.asarray(L)
        W = np.asarray(W)

        npts = len(H)
        n_batches = (npts + batch_size - 1) // batch_size

        results = []
        for ib in range(n_batches):
            start = ib * batch_size
            end = min((ib + 1) * batch_size, npts)

            H_batch = H[start:end]
            K_batch = K[start:end]
            L_batch = L[start:end]
            W_batch = W[start:end]
            EXP_batch = EXP[start:end] if len(EXP) > 1 else EXP

            # Update lattice for this batch
            self.resolution.lattice.npts = end - start

            conv_batch = self.convolve(sqw_func, None, H_batch, K_batch, L_batch,
                                       W_batch, EXP_batch, p, method, accuracy)
            results.append(conv_batch)

        return np.concatenate(results)
